import collections
import sys
import unittest

import example_robot_data
import numpy as np
import pinocchio
from factory import Contact1DModelDerived, Contact3DModelDerived, Contact6DModelDerived

import crocoddyl


class ContactModelAbstractTestCase(unittest.TestCase):
    ROBOT_MODEL = None
    ROBOT_STATE = None
    CONTACT = None
    CONTACT_DER = None

    def setUp(self):
        self.x = self.ROBOT_STATE.rand()
        self.robot_data = self.ROBOT_MODEL.createData()
        self.data = self.CONTACT.createData(self.robot_data)
        self.data_der = self.CONTACT_DER.createData(self.robot_data)

        nq, nv = self.ROBOT_MODEL.nq, self.ROBOT_MODEL.nv
        pinocchio.forwardKinematics(
            self.ROBOT_MODEL,
            self.robot_data,
            self.x[:nq],
            self.x[nq:],
            pinocchio.utils.zero(nv),
        )
        pinocchio.computeJointJacobians(self.ROBOT_MODEL, self.robot_data)
        pinocchio.updateFramePlacements(self.ROBOT_MODEL, self.robot_data)
        pinocchio.computeForwardKinematicsDerivatives(
            self.ROBOT_MODEL,
            self.robot_data,
            self.x[:nq],
            self.x[nq:],
            pinocchio.utils.zero(nv),
        )

    def test_nc_dimension(self):
        self.assertEqual(self.CONTACT.nc, self.CONTACT_DER.nc, "Wrong nc.")

    def test_calc(self):
        # Run calc for both action models
        self.CONTACT.calc(self.data, self.x)
        self.CONTACT_DER.calc(self.data_der, self.x)
        # Checking the cost value and its residual
        self.assertTrue(
            np.allclose(self.data.Jc, self.data_der.Jc, atol=1e-9),
            "Wrong contact Jacobian (Jc).",
        )
        self.assertTrue(
            np.allclose(self.data.a0, self.data_der.a0, atol=1e-9),
            "Wrong drift acceleration (a0).",
        )

    def test_calcDiff(self):
        # Run calcDiff for both action models
        self.CONTACT.calc(self.data, self.x)
        self.CONTACT.calcDiff(self.data, self.x)
        self.CONTACT_DER.calc(self.data_der, self.x)
        self.CONTACT_DER.calcDiff(self.data_der, self.x)
        # Checking the Jacobians of the contact constraint
        self.assertTrue(
            np.allclose(self.data.da0_dx, self.data_der.da0_dx, atol=1e-9),
            "Wrong derivatives of the desired contact acceleration (da0_dx).",
        )

    def test_updateForce(self):
        # Run updateForce for both action models
        self.CONTACT.calc(self.data, self.x)
        self.CONTACT_DER.calc(self.data_der, self.x)
        rng = np.random.default_rng()
        force = rng.random(self.CONTACT.nc)
        self.CONTACT.updateForce(self.data, force)
        self.CONTACT_DER.updateForce(self.data_der, force)
        self.assertTrue(
            np.allclose(self.data.fext.vector, self.data_der.fext.vector, atol=1e-9),
            "Wrong external spatial force.",
        )


class ContactModelMultipleAbstractTestCase(unittest.TestCase):
    ROBOT_MODEL = None
    ROBOT_STATE = None
    CONTACTS = None

    def setUp(self):
        self.x = self.ROBOT_STATE.rand()
        self.robot_data = self.ROBOT_MODEL.createData()

        self.contactSum = crocoddyl.ContactModelMultiple(self.ROBOT_STATE)
        self.datas = collections.OrderedDict(
            [
                [name, contact.createData(self.robot_data)]
                for name, contact in self.CONTACTS.items()
            ]
        )
        for name, contact in self.CONTACTS.items():
            self.contactSum.addContact(name, contact)
        self.dataSum = self.contactSum.createData(self.robot_data)

        nq, nv = self.ROBOT_MODEL.nq, self.ROBOT_MODEL.nv
        pinocchio.forwardKinematics(
            self.ROBOT_MODEL,
            self.robot_data,
            self.x[:nq],
            self.x[nq:],
            pinocchio.utils.zero(nv),
        )
        pinocchio.computeJointJacobians(self.ROBOT_MODEL, self.robot_data)
        pinocchio.updateFramePlacements(self.ROBOT_MODEL, self.robot_data)
        pinocchio.computeForwardKinematicsDerivatives(
            self.ROBOT_MODEL,
            self.robot_data,
            self.x[:nq],
            self.x[nq:],
            pinocchio.utils.zero(nv),
        )

    def test_nc_dimension(self):
        nc = sum([contact.nc for contact in self.CONTACTS.values()])
        self.assertEqual(self.contactSum.nc, nc, "Wrong nc.")

    def test_calc(self):
        # Run calc for both action models
        for contact, data in zip(self.CONTACTS.values(), self.datas.values()):
            contact.calc(data, self.x)
        self.contactSum.calc(self.dataSum, self.x)
        # Checking the cost value and its residual
        Jc = np.vstack([data.Jc for data in self.datas.values()])
        a0 = np.hstack([data.a0 for data in self.datas.values()])
        self.assertTrue(
            np.allclose(self.dataSum.Jc, Jc, atol=1e-9), "Wrong contact Jacobian (Jc)."
        )
        self.assertTrue(
            np.allclose(self.dataSum.a0, a0, atol=1e-9),
            "Wrong drift acceleration (a0).",
        )

    def test_calcDiff(self):
        # Run calc for both action models
        for contact, data in zip(self.CONTACTS.values(), self.datas.values()):
            contact.calc(data, self.x)
            contact.calcDiff(data, self.x)
        self.contactSum.calc(self.dataSum, self.x)
        self.contactSum.calcDiff(self.dataSum, self.x)
        # Checking the Jacobians of the contact constraint
        da0_dx = np.vstack([data.da0_dx for data in self.datas.values()])
        self.assertTrue(
            np.allclose(self.dataSum.da0_dx, da0_dx, atol=1e-9),
            "Wrong derivatives of the desired contact acceleration (da0_dx).",
        )


class Contact1DLocalTest(ContactModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("hyq").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    gains = pinocchio.utils.rand(2)
    xref = pinocchio.SE3.Random().translation[2]
    CONTACT = crocoddyl.ContactModel1D(
        ROBOT_STATE, ROBOT_MODEL.getFrameId("lf_foot"), xref, pinocchio.LOCAL, gains
    )
    CONTACT_DER = Contact1DModelDerived(
        ROBOT_STATE, ROBOT_MODEL.getFrameId("lf_foot"), xref, pinocchio.LOCAL, gains
    )
    Raxis = pinocchio.SE3.Random().rotation
    CONTACT.Raxis = Raxis
    CONTACT_DER.Raxis = Raxis


class Contact1DWorldTest(ContactModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("hyq").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    gains = pinocchio.utils.rand(2)
    xref = pinocchio.SE3.Random().translation[2]
    CONTACT = crocoddyl.ContactModel1D(
        ROBOT_STATE, ROBOT_MODEL.getFrameId("lf_foot"), xref, pinocchio.WORLD, gains
    )
    CONTACT_DER = Contact1DModelDerived(
        ROBOT_STATE, ROBOT_MODEL.getFrameId("lf_foot"), xref, pinocchio.WORLD, gains
    )
    Raxis = pinocchio.SE3.Random().rotation
    CONTACT.Raxis = Raxis
    CONTACT_DER.Raxis = Raxis


class Contact1DLocalWorldAlignedTest(ContactModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("hyq").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    gains = pinocchio.utils.rand(2)
    xref = pinocchio.SE3.Random().translation[2]
    CONTACT = crocoddyl.ContactModel1D(
        ROBOT_STATE,
        ROBOT_MODEL.getFrameId("lf_foot"),
        xref,
        pinocchio.LOCAL_WORLD_ALIGNED,
        gains,
    )
    CONTACT_DER = Contact1DModelDerived(
        ROBOT_STATE,
        ROBOT_MODEL.getFrameId("lf_foot"),
        xref,
        pinocchio.LOCAL_WORLD_ALIGNED,
        gains,
    )
    Raxis = pinocchio.SE3.Random().rotation
    CONTACT.Raxis = Raxis
    CONTACT_DER.Raxis = Raxis


class Contact3DLocalTest(ContactModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("hyq").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    gains = pinocchio.utils.rand(2)
    xref = pinocchio.SE3.Random().translation
    CONTACT = crocoddyl.ContactModel3D(
        ROBOT_STATE, ROBOT_MODEL.getFrameId("lf_foot"), xref, pinocchio.LOCAL, gains
    )
    CONTACT_DER = Contact3DModelDerived(
        ROBOT_STATE, ROBOT_MODEL.getFrameId("lf_foot"), xref, pinocchio.LOCAL, gains
    )


class Contact3DWorldTest(ContactModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("hyq").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    gains = pinocchio.utils.rand(2)
    xref = pinocchio.SE3.Random().translation
    CONTACT = crocoddyl.ContactModel3D(
        ROBOT_STATE, ROBOT_MODEL.getFrameId("lf_foot"), xref, pinocchio.WORLD, gains
    )
    CONTACT_DER = Contact3DModelDerived(
        ROBOT_STATE, ROBOT_MODEL.getFrameId("lf_foot"), xref, pinocchio.WORLD, gains
    )


class Contact3DLocalWorldAlignedTest(ContactModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("hyq").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    gains = pinocchio.utils.rand(2)
    xref = pinocchio.SE3.Random().translation
    CONTACT = crocoddyl.ContactModel3D(
        ROBOT_STATE,
        ROBOT_MODEL.getFrameId("lf_foot"),
        xref,
        pinocchio.LOCAL_WORLD_ALIGNED,
        gains,
    )
    CONTACT_DER = Contact3DModelDerived(
        ROBOT_STATE,
        ROBOT_MODEL.getFrameId("lf_foot"),
        xref,
        pinocchio.LOCAL_WORLD_ALIGNED,
        gains,
    )


class Contact3DMultipleTest(ContactModelMultipleAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("hyq").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    gains = pinocchio.utils.rand(2)
    CONTACTS = collections.OrderedDict(
        sorted(
            {
                "lf_foot": crocoddyl.ContactModel3D(
                    ROBOT_STATE,
                    ROBOT_MODEL.getFrameId("lf_foot"),
                    pinocchio.SE3.Random().translation,
                    pinocchio.LOCAL,
                    gains,
                ),
                "rh_foot": crocoddyl.ContactModel3D(
                    ROBOT_STATE,
                    ROBOT_MODEL.getFrameId("rh_foot"),
                    pinocchio.SE3.Random().translation,
                    pinocchio.LOCAL,
                    gains,
                ),
            }.items()
        )
    )


class Contact6DLocalTest(ContactModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("icub_reduced").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    gains = pinocchio.utils.rand(2)
    Mref = pinocchio.SE3.Random()
    CONTACT = crocoddyl.ContactModel6D(
        ROBOT_STATE, ROBOT_MODEL.getFrameId("r_sole"), Mref, pinocchio.LOCAL, gains
    )
    CONTACT_DER = Contact6DModelDerived(
        ROBOT_STATE, ROBOT_MODEL.getFrameId("r_sole"), Mref, pinocchio.LOCAL, gains
    )


class Contact6DWorldTest(ContactModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("icub_reduced").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    gains = pinocchio.utils.rand(2)
    Mref = pinocchio.SE3.Random()
    CONTACT = crocoddyl.ContactModel6D(
        ROBOT_STATE, ROBOT_MODEL.getFrameId("r_sole"), Mref, pinocchio.WORLD, gains
    )
    CONTACT_DER = Contact6DModelDerived(
        ROBOT_STATE, ROBOT_MODEL.getFrameId("r_sole"), Mref, pinocchio.WORLD, gains
    )


class Contact6DLocalWorldAlignedTest(ContactModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("icub_reduced").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    gains = pinocchio.utils.rand(2)
    Mref = pinocchio.SE3.Random()
    CONTACT = crocoddyl.ContactModel6D(
        ROBOT_STATE,
        ROBOT_MODEL.getFrameId("r_sole"),
        Mref,
        pinocchio.LOCAL_WORLD_ALIGNED,
        gains,
    )
    CONTACT_DER = Contact6DModelDerived(
        ROBOT_STATE,
        ROBOT_MODEL.getFrameId("r_sole"),
        Mref,
        pinocchio.LOCAL_WORLD_ALIGNED,
        gains,
    )


class Contact6DMultipleTest(ContactModelMultipleAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("icub_reduced").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    gains = pinocchio.utils.rand(2)
    CONTACTS = collections.OrderedDict(
        sorted(
            {
                "l_foot": crocoddyl.ContactModel6D(
                    ROBOT_STATE,
                    ROBOT_MODEL.getFrameId("l_sole"),
                    pinocchio.SE3.Random(),
                    pinocchio.LOCAL,
                    gains,
                ),
                "r_foot": crocoddyl.ContactModel6D(
                    ROBOT_STATE,
                    ROBOT_MODEL.getFrameId("r_sole"),
                    pinocchio.SE3.Random(),
                    pinocchio.LOCAL,
                    gains,
                ),
            }.items()
        )
    )


if __name__ == "__main__":
    # test to be run
    test_classes_to_run = [
        Contact1DLocalTest,
        Contact1DWorldTest,
        Contact1DLocalWorldAlignedTest,
        Contact3DLocalTest,
        Contact3DWorldTest,
        Contact3DLocalWorldAlignedTest,
        Contact3DMultipleTest,
        Contact6DLocalTest,
        Contact6DWorldTest,
        Contact6DLocalWorldAlignedTest,
        Contact6DMultipleTest,
    ]
    loader = unittest.TestLoader()
    suites_list = []
    for test_class in test_classes_to_run:
        suite = loader.loadTestsFromTestCase(test_class)
        suites_list.append(suite)
    big_suite = unittest.TestSuite(suites_list)
    runner = unittest.TextTestRunner()
    results = runner.run(big_suite)
    sys.exit(not results.wasSuccessful())



import sys
import unittest
from random import randint

import example_robot_data
import numpy as np
import pinocchio
from factory import (
    DifferentialFreeFwdDynamicsModelDerived,
    DifferentialLQRModelDerived,
    IntegratedActionModelEulerDerived,
    IntegratedActionModelRK4Derived,
    LQRModelDerived,
    UnicycleModelDerived,
)

import crocoddyl


class ActionModelAbstractTestCase(unittest.TestCase):
    MODEL = None
    MODEL_DER = None

    def setUp(self):
        rng = np.random.default_rng()
        state = self.MODEL.state
        self.x = state.rand()
        self.u = rng.random(self.MODEL.nu)
        self.DATA = self.MODEL.createData()
        self.DATA_DER = self.MODEL_DER.createData()

    def test_calc(self):
        # Run calc for both action models
        self.MODEL.calc(self.DATA, self.x, self.u)
        self.MODEL_DER.calc(self.DATA_DER, self.x, self.u)
        # Checking the cost value and its residual
        self.assertAlmostEqual(
            self.DATA.cost, self.DATA_DER.cost, 10, "Wrong cost value."
        )
        self.assertTrue(
            np.allclose(self.DATA.r, self.DATA_DER.r, atol=1e-9),
            "Wrong cost residuals.",
        )

        if isinstance(self.MODEL, crocoddyl.ActionModelAbstract):
            # Checking the dimension of the next state
            self.assertEqual(
                self.DATA.xnext.shape,
                self.DATA_DER.xnext.shape,
                "Wrong next state dimension.",
            )
            # Checking the next state value
            self.assertTrue(
                np.allclose(self.DATA.xnext, self.DATA_DER.xnext, atol=1e-9),
                "Wrong next state.",
            )
        elif isinstance(self.MODEL, crocoddyl.DifferentialActionModelAbstract):
            # Checking the dimension of the next state
            self.assertEqual(
                self.DATA.xout.shape,
                self.DATA_DER.xout.shape,
                "Wrong next state dimension.",
            )
            # Checking the next state value
            self.assertTrue(
                np.allclose(self.DATA.xout, self.DATA_DER.xout, atol=1e-9),
                "Wrong next state.",
            )

    def test_calc_x(self):
        # Run calc for both action models
        self.MODEL.calc(self.DATA, self.x)
        self.MODEL_DER.calc(self.DATA_DER, self.x)
        # Checking the cost value and its residual
        self.assertAlmostEqual(
            self.DATA.cost, self.DATA_DER.cost, 10, "Wrong cost value."
        )
        self.assertTrue(
            np.allclose(self.DATA.r, self.DATA_DER.r, atol=1e-9),
            "Wrong cost residuals.",
        )

        if isinstance(self.MODEL, crocoddyl.ActionModelAbstract):
            # Checking the dimension of the next state
            self.assertEqual(
                self.DATA.xnext.shape,
                self.DATA_DER.xnext.shape,
                "Wrong next state dimension.",
            )
            # Checking the next state value
            self.assertTrue(
                np.allclose(self.DATA.xnext, self.DATA_DER.xnext, atol=1e-9),
                "Wrong next state.",
            )
        elif isinstance(self.MODEL, crocoddyl.DifferentialActionModelAbstract):
            # Checking the dimension of the next state
            self.assertEqual(
                self.DATA.xout.shape,
                self.DATA_DER.xout.shape,
                "Wrong next state dimension.",
            )
            # Checking the next state value
            self.assertTrue(
                np.allclose(self.DATA.xout, self.DATA_DER.xout, atol=1e-9),
                "Wrong next state.",
            )

    def test_calcDiff(self):
        # Run calcDiff for both action models
        self.MODEL.calc(self.DATA, self.x, self.u)
        self.MODEL.calcDiff(self.DATA, self.x, self.u)

        self.MODEL_DER.calc(self.DATA_DER, self.x, self.u)
        self.MODEL_DER.calcDiff(self.DATA_DER, self.x, self.u)
        # Checking the next state value
        if isinstance(self.MODEL, crocoddyl.ActionModelAbstract):
            self.assertTrue(
                np.allclose(self.DATA.xnext, self.DATA_DER.xnext, atol=1e-9),
                "Wrong next state.",
            )
        elif isinstance(self.MODEL, crocoddyl.DifferentialActionModelAbstract):
            self.assertTrue(
                np.allclose(self.DATA.xout, self.DATA_DER.xout, atol=1e-9),
                "Wrong next state.",
            )
        # Checking the Jacobians of the dynamic
        self.assertTrue(
            np.allclose(self.DATA.Fx, self.DATA_DER.Fx, atol=1e-9), "Wrong Fx."
        )
        self.assertTrue(
            np.allclose(self.DATA.Fu, self.DATA_DER.Fu, atol=1e-9), "Wrong Fu."
        )
        # Checking the Jacobians and Hessians of the cost
        self.assertTrue(
            np.allclose(self.DATA.Lx, self.DATA_DER.Lx, atol=1e-9), "Wrong Lx."
        )
        self.assertTrue(
            np.allclose(self.DATA.Lu, self.DATA_DER.Lu, atol=1e-9), "Wrong Lu."
        )
        self.assertTrue(
            np.allclose(self.DATA.Lxx, self.DATA_DER.Lxx, atol=1e-9), "Wrong Lxx."
        )
        self.assertTrue(
            np.allclose(self.DATA.Lxu, self.DATA_DER.Lxu, atol=1e-9), "Wrong Lxu."
        )
        self.assertTrue(
            np.allclose(self.DATA.Luu, self.DATA_DER.Luu, atol=1e-9), "Wrong Luu."
        )

    def test_calcDiff_x(self):
        # Run calcDiff for both action models
        self.MODEL.calc(self.DATA, self.x)
        self.MODEL.calcDiff(self.DATA, self.x)

        self.MODEL_DER.calc(self.DATA_DER, self.x)
        self.MODEL_DER.calcDiff(self.DATA_DER, self.x)
        # Checking the next state value
        if isinstance(self.MODEL, crocoddyl.ActionModelAbstract):
            self.assertTrue(
                np.allclose(self.DATA.xnext, self.DATA_DER.xnext, atol=1e-9),
                "Wrong next state.",
            )
        # Checking the Jacobians and Hessians of the cost
        self.assertTrue(
            np.allclose(self.DATA.Lx, self.DATA_DER.Lx, atol=1e-9), "Wrong Lx."
        )
        self.assertTrue(
            np.allclose(self.DATA.Lxx, self.DATA_DER.Lxx, atol=1e-9), "Wrong Lxx."
        )


class UnicycleTest(ActionModelAbstractTestCase):
    MODEL = crocoddyl.ActionModelUnicycle()
    MODEL_DER = UnicycleModelDerived()


class LQRTest(ActionModelAbstractTestCase):
    NX = randint(2, 21)
    NU = randint(2, NX)
    MODEL = crocoddyl.ActionModelLQR(NX, NU)
    MODEL_DER = LQRModelDerived(NX, NU)


class DifferentialLQRTest(ActionModelAbstractTestCase):
    NX = randint(2, 21)
    NU = randint(2, NX)
    MODEL = crocoddyl.DifferentialActionModelLQR(NX, NU)
    MODEL_DER = DifferentialLQRModelDerived(NX, NU)


class TalosArmFreeFwdDynamicsTest(ActionModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("talos_arm").model
    STATE = crocoddyl.StateMultibody(ROBOT_MODEL)
    ACTUATION = crocoddyl.ActuationModelFull(STATE)
    COST_SUM = crocoddyl.CostModelSum(STATE)
    COST_SUM.addCost(
        "gripperPose",
        crocoddyl.CostModelResidual(
            STATE,
            crocoddyl.ResidualModelFramePlacement(
                STATE,
                ROBOT_MODEL.getFrameId("gripper_left_joint"),
                pinocchio.SE3.Random(),
            ),
        ),
        1e-3,
    )
    COST_SUM.addCost(
        "xReg",
        crocoddyl.CostModelResidual(STATE, crocoddyl.ResidualModelState(STATE)),
        1e-7,
    )
    COST_SUM.addCost(
        "uReg",
        crocoddyl.CostModelResidual(STATE, crocoddyl.ResidualModelControl(STATE)),
        1e-7,
    )
    MODEL = crocoddyl.DifferentialActionModelFreeFwdDynamics(STATE, ACTUATION, COST_SUM)
    MODEL_DER = DifferentialFreeFwdDynamicsModelDerived(STATE, ACTUATION, COST_SUM)


class TalosArmFreeFwdDynamicsWithArmatureTest(ActionModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("talos_arm").model
    STATE = crocoddyl.StateMultibody(ROBOT_MODEL)
    ACTUATION = crocoddyl.ActuationModelFull(STATE)
    COST_SUM = crocoddyl.CostModelSum(STATE)
    COST_SUM.addCost(
        "gripperPose",
        crocoddyl.CostModelResidual(
            STATE,
            crocoddyl.ResidualModelFramePlacement(
                STATE,
                ROBOT_MODEL.getFrameId("gripper_left_joint"),
                pinocchio.SE3.Random(),
            ),
        ),
        1e-3,
    )
    COST_SUM.addCost(
        "xReg",
        crocoddyl.CostModelResidual(STATE, crocoddyl.ResidualModelState(STATE)),
        1e-7,
    )
    COST_SUM.addCost(
        "uReg",
        crocoddyl.CostModelResidual(STATE, crocoddyl.ResidualModelControl(STATE)),
        1e-7,
    )
    MODEL = crocoddyl.DifferentialActionModelFreeFwdDynamics(STATE, ACTUATION, COST_SUM)
    MODEL_DER = DifferentialFreeFwdDynamicsModelDerived(STATE, ACTUATION, COST_SUM)
    MODEL.armature = 0.1 * np.ones(ROBOT_MODEL.nv)
    MODEL_DER.set_armature(0.1 * np.ones(ROBOT_MODEL.nv))


class AnymalFreeFwdDynamicsTest(ActionModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("anymal").model
    STATE = crocoddyl.StateMultibody(ROBOT_MODEL)
    ACTUATION = crocoddyl.ActuationModelFloatingBase(STATE)
    COST_SUM = crocoddyl.CostModelSum(STATE, ACTUATION.nu)
    COST_SUM.addCost(
        "xReg",
        crocoddyl.CostModelResidual(
            STATE, crocoddyl.ResidualModelState(STATE, ACTUATION.nu)
        ),
        1e-7,
    )
    COST_SUM.addCost(
        "uReg",
        crocoddyl.CostModelResidual(
            STATE, crocoddyl.ResidualModelControl(STATE, ACTUATION.nu)
        ),
        1e-7,
    )
    MODEL = crocoddyl.DifferentialActionModelFreeFwdDynamics(STATE, ACTUATION, COST_SUM)
    MODEL_DER = DifferentialFreeFwdDynamicsModelDerived(STATE, ACTUATION, COST_SUM)


class TalosArmIntegratedEulerTest(ActionModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("talos_arm").model
    STATE = crocoddyl.StateMultibody(ROBOT_MODEL)
    ACTUATION = crocoddyl.ActuationModelFull(STATE)
    COST_SUM = crocoddyl.CostModelSum(STATE)
    COST_SUM.addCost(
        "gripperPose",
        crocoddyl.CostModelResidual(
            STATE,
            crocoddyl.ResidualModelFramePlacement(
                STATE,
                ROBOT_MODEL.getFrameId("gripper_left_joint"),
                pinocchio.SE3.Random(),
            ),
        ),
        1e-3,
    )
    COST_SUM.addCost(
        "xReg",
        crocoddyl.CostModelResidual(STATE, crocoddyl.ResidualModelState(STATE)),
        1e-7,
    )
    COST_SUM.addCost(
        "uReg",
        crocoddyl.CostModelResidual(STATE, crocoddyl.ResidualModelControl(STATE)),
        1e-7,
    )
    DIFFERENTIAL = crocoddyl.DifferentialActionModelFreeFwdDynamics(
        STATE, ACTUATION, COST_SUM
    )
    MODEL = crocoddyl.IntegratedActionModelEuler(DIFFERENTIAL, 1e-3)
    MODEL_DER = IntegratedActionModelEulerDerived(DIFFERENTIAL, 1e-3)


class TalosArmIntegratedRK4Test(ActionModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("talos_arm").model
    STATE = crocoddyl.StateMultibody(ROBOT_MODEL)
    ACTUATION = crocoddyl.ActuationModelFull(STATE)
    COST_SUM = crocoddyl.CostModelSum(STATE)
    COST_SUM.addCost(
        "gripperPose",
        crocoddyl.CostModelResidual(
            STATE,
            crocoddyl.ResidualModelFramePlacement(
                STATE,
                ROBOT_MODEL.getFrameId("gripper_left_joint"),
                pinocchio.SE3.Random(),
            ),
        ),
        1e-3,
    )
    COST_SUM.addCost(
        "xReg",
        crocoddyl.CostModelResidual(STATE, crocoddyl.ResidualModelState(STATE)),
        1e-7,
    )
    COST_SUM.addCost(
        "uReg",
        crocoddyl.CostModelResidual(STATE, crocoddyl.ResidualModelControl(STATE)),
        1e-7,
    )
    DIFFERENTIAL = crocoddyl.DifferentialActionModelFreeFwdDynamics(
        STATE, ACTUATION, COST_SUM
    )
    MODEL = crocoddyl.IntegratedActionModelRK(DIFFERENTIAL, crocoddyl.RKType.four, 1e-3)
    MODEL_DER = IntegratedActionModelRK4Derived(DIFFERENTIAL, 1e-3)


class AnymalIntegratedEulerTest(ActionModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("anymal").model
    STATE = crocoddyl.StateMultibody(ROBOT_MODEL)
    ACTUATION = crocoddyl.ActuationModelFloatingBase(STATE)
    COST_SUM = crocoddyl.CostModelSum(STATE, ACTUATION.nu)
    COST_SUM.addCost(
        "xReg",
        crocoddyl.CostModelResidual(
            STATE, crocoddyl.ResidualModelState(STATE, ACTUATION.nu)
        ),
        1e-7,
    )
    COST_SUM.addCost(
        "uReg",
        crocoddyl.CostModelResidual(
            STATE, crocoddyl.ResidualModelControl(STATE, ACTUATION.nu)
        ),
        1e-7,
    )
    DIFFERENTIAL = crocoddyl.DifferentialActionModelFreeFwdDynamics(
        STATE, ACTUATION, COST_SUM
    )
    MODEL = crocoddyl.IntegratedActionModelRK(DIFFERENTIAL, crocoddyl.RKType.four, 1e-3)
    MODEL_DER = IntegratedActionModelRK4Derived(DIFFERENTIAL, 1e-3)


class AnymalIntegratedRK4Test(ActionModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("anymal").model
    STATE = crocoddyl.StateMultibody(ROBOT_MODEL)
    ACTUATION = crocoddyl.ActuationModelFloatingBase(STATE)
    COST_SUM = crocoddyl.CostModelSum(STATE, ACTUATION.nu)
    COST_SUM.addCost(
        "xReg",
        crocoddyl.CostModelResidual(
            STATE, crocoddyl.ResidualModelState(STATE, ACTUATION.nu)
        ),
        1e-7,
    )
    COST_SUM.addCost(
        "uReg",
        crocoddyl.CostModelResidual(
            STATE, crocoddyl.ResidualModelControl(STATE, ACTUATION.nu)
        ),
        1e-7,
    )
    DIFFERENTIAL = crocoddyl.DifferentialActionModelFreeFwdDynamics(
        STATE, ACTUATION, COST_SUM
    )
    MODEL = crocoddyl.IntegratedActionModelRK(DIFFERENTIAL, crocoddyl.RKType.four, 1e-3)
    MODEL_DER = IntegratedActionModelRK4Derived(DIFFERENTIAL, 1e-3)


if __name__ == "__main__":
    # test to be run
    test_classes_to_run = [
        UnicycleTest,
        LQRTest,
        DifferentialLQRTest,
        TalosArmFreeFwdDynamicsTest,
        TalosArmFreeFwdDynamicsWithArmatureTest,
        AnymalFreeFwdDynamicsTest,
        AnymalIntegratedEulerTest,
        AnymalIntegratedRK4Test,
        TalosArmIntegratedRK4Test,
        TalosArmIntegratedEulerTest,
    ]
    loader = unittest.TestLoader()
    suites_list = []
    for test_class in test_classes_to_run:
        suite = loader.loadTestsFromTestCase(test_class)
        suites_list.append(suite)
    big_suite = unittest.TestSuite(suites_list)
    runner = unittest.TextTestRunner()
    results = runner.run(big_suite)
    sys.exit(not results.wasSuccessful())



import sys
import unittest
from random import randint

import example_robot_data
import numpy as np
import pinocchio
from factory import DDPDerived, FDDPDerived

import crocoddyl


class SolverAbstractTestCase(unittest.TestCase):
    MODEL = None
    SOLVER = None
    SOLVER_DER = None

    def setUp(self):
        # Set up the solvers
        self.T = randint(1, 21)
        state = self.MODEL.state
        self.xs = []
        self.us = []
        self.xs.append(state.rand())
        rng = np.random.default_rng()
        for _ in range(self.T):
            self.xs.append(state.rand())
            self.us.append(rng.random(self.MODEL.nu))
        self.PROBLEM = crocoddyl.ShootingProblem(
            self.xs[0], [self.MODEL] * self.T, self.MODEL
        )
        self.PROBLEM_DER = crocoddyl.ShootingProblem(
            self.xs[0], [self.MODEL] * self.T, self.MODEL
        )
        self.solver = self.SOLVER(self.PROBLEM)
        self.solver_der = self.SOLVER_DER(self.PROBLEM_DER)

    def test_number_of_nodes(self):
        # Check the number of nodes
        self.assertEqual(
            self.T, self.solver.problem.T, "Wrong number of nodes in SOLVER"
        )
        self.assertEqual(
            self.T, self.solver_der.problem.T, "Wrong number of nodes in SOLVER_DER"
        )

    def test_solve(self):
        # Run maximum 10 iterations in order to boost test analysis
        self.solver.solve([], [], 10)
        self.solver_der.solve([], [], 10)
        for x1, x2 in zip(self.solver.xs, self.solver_der.xs):
            self.assertTrue(np.allclose(x1, x2, atol=1e-9), "xs doesn't match.")
        for u1, u2 in zip(self.solver.us, self.solver_der.us):
            self.assertTrue(np.allclose(u1, u2, atol=1e-9), "us doesn't match.")
        for k1, k2 in zip(self.solver.k, self.solver_der.k):
            self.assertTrue(np.allclose(k1, k2, atol=1e-9), "k doesn't match.")

    def test_compute_search_direction(self):
        # Compute the direction
        self.solver.setCandidate([], [], False)
        self.solver_der.setCandidate([], [], False)
        self.solver.computeDirection()
        self.solver_der.computeDirection()
        # Check the LQ model of the Hamiltonian
        for qx1, qx2 in zip(self.solver.Qx, self.solver_der.Qx):
            self.assertTrue(np.allclose(qx1, qx2, atol=1e-9), "Qx doesn't match.")
        for qu1, qu2 in zip(self.solver.Qu, self.solver_der.Qu):
            self.assertTrue(np.allclose(qu1, qu2, atol=1e-9), "Qu doesn't match.")
        for qxx1, qxx2 in zip(self.solver.Qxx, self.solver_der.Qxx):
            self.assertTrue(np.allclose(qxx1, qxx2, atol=1e-9), "Qxx doesn't match.")
        for qxu1, qxu2 in zip(self.solver.Qxu, self.solver_der.Qxu):
            self.assertTrue(np.allclose(qxu1, qxu2, atol=1e-9), "Quu doesn't match.")
        for quu1, quu2 in zip(self.solver.Quu, self.solver_der.Quu):
            self.assertTrue(np.allclose(quu1, quu2, atol=1e-9), "Quu doesn't match.")
        for vx1, vx2 in zip(self.solver.Vx, self.solver_der.Vx):
            self.assertTrue(np.allclose(vx1, vx2, atol=1e-9), "Vx doesn't match.")
        for vxx1, vxx2 in zip(self.solver.Vxx, self.solver_der.Vxx):
            self.assertTrue(np.allclose(vxx1, vxx2, atol=1e-9), "Vxx doesn't match.")

    def test_try_step(self):
        # Try a full step and check the improvement in the cost
        self.solver.setCandidate([], [], False)
        self.solver_der.setCandidate([], [], False)
        self.solver.computeDirection()
        self.solver_der.computeDirection()
        cost = self.solver.tryStep()
        costDer = self.solver_der.tryStep()
        self.assertAlmostEqual(cost, costDer, 9, "Wrong cost value for full step")
        # Try a half step and check the improvement in the cost
        cost = self.solver.tryStep(0.5)
        costDer = self.solver_der.tryStep(0.5)
        self.assertAlmostEqual(cost, costDer, 9, "Wrong cost value for half step")

    def test_stopping_criteria(self):
        # Run 2 iteration in order to boost test analysis
        self.solver.solve([], [], 2)
        self.solver_der.solve([], [], 2)
        # Compute and check the stopping criteria
        stop = self.solver.stoppingCriteria()
        stopDer = self.solver_der.stoppingCriteria()
        self.assertAlmostEqual(stop, stopDer, 9, "Wrong stopping value")

    def test_expected_improvement(self):
        # Run 2 iteration in order to boost test analysis
        self.solver.solve([], [], 2)
        self.solver_der.solve([], [], 2)
        expImp = self.solver.expectedImprovement()
        expImpDer = self.solver_der.expectedImprovement()
        self.assertTrue(
            np.allclose(expImp, expImpDer, atol=1e-9),
            "Expected improvement doesn't match.",
        )


class UnicycleDDPTest(SolverAbstractTestCase):
    MODEL = crocoddyl.ActionModelUnicycle()
    SOLVER = crocoddyl.SolverDDP
    SOLVER_DER = DDPDerived


class UnicycleFDDPTest(SolverAbstractTestCase):
    MODEL = crocoddyl.ActionModelUnicycle()
    SOLVER = crocoddyl.SolverFDDP
    SOLVER_DER = FDDPDerived


class TalosArmDDPTest(SolverAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("talos_arm").model
    STATE = crocoddyl.StateMultibody(ROBOT_MODEL)
    ACTUATION = crocoddyl.ActuationModelFull(STATE)
    COST_SUM = crocoddyl.CostModelSum(STATE)
    COST_SUM.addCost(
        "gripperPose",
        crocoddyl.CostModelResidual(
            STATE,
            crocoddyl.ResidualModelFramePlacement(
                STATE,
                ROBOT_MODEL.getFrameId("gripper_left_joint"),
                pinocchio.SE3.Random(),
            ),
        ),
        1e-5,
    )
    COST_SUM.addCost(
        "xReg",
        crocoddyl.CostModelResidual(STATE, crocoddyl.ResidualModelState(STATE)),
        1e-7,
    )
    COST_SUM.addCost(
        "uReg",
        crocoddyl.CostModelResidual(STATE, crocoddyl.ResidualModelControl(STATE)),
        1e-7,
    )
    DIFF_MODEL = crocoddyl.DifferentialActionModelFreeFwdDynamics(
        STATE, ACTUATION, COST_SUM
    )
    MODEL = crocoddyl.IntegratedActionModelEuler(DIFF_MODEL, 1e-3)
    SOLVER = crocoddyl.SolverDDP
    SOLVER_DER = DDPDerived


class TalosArmFDDPTest(SolverAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("talos_arm").model
    STATE = crocoddyl.StateMultibody(ROBOT_MODEL)
    ACTUATION = crocoddyl.ActuationModelFull(STATE)
    COST_SUM = crocoddyl.CostModelSum(STATE)
    COST_SUM.addCost(
        "gripperPose",
        crocoddyl.CostModelResidual(
            STATE,
            crocoddyl.ResidualModelFramePlacement(
                STATE,
                ROBOT_MODEL.getFrameId("gripper_left_joint"),
                pinocchio.SE3.Random(),
            ),
        ),
        1e-5,
    )
    COST_SUM.addCost(
        "xReg",
        crocoddyl.CostModelResidual(STATE, crocoddyl.ResidualModelState(STATE)),
        1e-7,
    )
    COST_SUM.addCost(
        "uReg",
        crocoddyl.CostModelResidual(STATE, crocoddyl.ResidualModelControl(STATE)),
        1e-7,
    )
    DIFF_MODEL = crocoddyl.DifferentialActionModelFreeFwdDynamics(
        STATE, ACTUATION, COST_SUM
    )
    MODEL = crocoddyl.IntegratedActionModelEuler(DIFF_MODEL, 1e-3)
    SOLVER = crocoddyl.SolverFDDP
    SOLVER_DER = FDDPDerived


if __name__ == "__main__":
    # test to be run
    test_classes_to_run = [
        UnicycleDDPTest,
        UnicycleFDDPTest,
        TalosArmDDPTest,
        TalosArmFDDPTest,
    ]
    loader = unittest.TestLoader()
    suites_list = []
    for test_class in test_classes_to_run:
        suite = loader.loadTestsFromTestCase(test_class)
        suites_list.append(suite)
    big_suite = unittest.TestSuite(suites_list)
    runner = unittest.TextTestRunner()
    results = runner.run(big_suite)
    sys.exit(not results.wasSuccessful())



import os
import signal
import sys
import time

import example_robot_data
import numpy as np
import pinocchio

import crocoddyl
from crocoddyl.utils.quadruped import SimpleQuadrupedalGaitProblem, plotSolution

WITHDISPLAY = "display" in sys.argv or "CROCODDYL_DISPLAY" in os.environ
WITHPLOT = "plot" in sys.argv or "CROCODDYL_PLOT" in os.environ
signal.signal(signal.SIGINT, signal.SIG_DFL)

# Loading the anymal model
anymal = example_robot_data.load("anymal")
lims = anymal.model.effortLimit
lims *= 0.4  # reduced artificially the torque limits
anymal.model.effortLimit = lims
lims = anymal.model.velocityLimit
lims *= 0.5
anymal.model.velocityLimit = lims

# Defining the initial state of the robot
q0 = anymal.model.referenceConfigurations["standing"].copy()
v0 = pinocchio.utils.zero(anymal.model.nv)
x0 = np.concatenate([q0, v0])

# Setting up the 3d walking problem
lfFoot, rfFoot, lhFoot, rhFoot = "LF_FOOT", "RF_FOOT", "LH_FOOT", "RH_FOOT"
gait = SimpleQuadrupedalGaitProblem(anymal.model, lfFoot, rfFoot, lhFoot, rhFoot)

# Defining the walking gait parameters
jumping_gait = {
    "jumpHeight": 0.15,
    "jumpLength": [0.3, 0.0, 0.0],
    "timeStep": 1e-2,
    "groundKnots": 20,
    "flyingKnots": 20,
}

# Creating a both solvers
boxfddp = crocoddyl.SolverBoxFDDP(
    gait.createJumpingProblem(
        x0,
        jumping_gait["jumpHeight"],
        jumping_gait["jumpLength"],
        jumping_gait["timeStep"],
        jumping_gait["groundKnots"],
        jumping_gait["flyingKnots"],
    )
)
boxddp = crocoddyl.SolverBoxDDP(
    gait.createJumpingProblem(
        x0,
        jumping_gait["jumpHeight"],
        jumping_gait["jumpLength"],
        jumping_gait["timeStep"],
        jumping_gait["groundKnots"],
        jumping_gait["flyingKnots"],
    )
)

# Added the callback functions
if WITHPLOT:
    boxfddp.setCallbacks(
        [
            crocoddyl.CallbackVerbose(),
            crocoddyl.CallbackLogger(),
        ]
    )
    boxddp.setCallbacks(
        [
            crocoddyl.CallbackVerbose(),
            crocoddyl.CallbackLogger(),
        ]
    )
else:
    boxfddp.setCallbacks([crocoddyl.CallbackVerbose()])
    boxddp.setCallbacks([crocoddyl.CallbackVerbose()])

# Solving the problem with the both solvers
xs = [x0] * (boxfddp.problem.T + 1)
us = boxddp.problem.quasiStatic([x0] * boxddp.problem.T)

print("*** SOLVE with Box-FDDP ***")
boxfddp.th_stop = 1e-7
boxfddp.solve(xs, us, 50, False)

print("*** SOLVE with Box-DDP ***")
boxddp.th_stop = 1e-7
boxddp.solve(xs, us, 30, False)

# Display the entire motion
if WITHDISPLAY:
    try:
        import gepetto

        gepetto.corbaserver.Client()
        cameraTF = [2.0, 2.68, 0.84, 0.2, 0.62, 0.72, 0.22]
        display = crocoddyl.GepettoDisplay(anymal, 4, 4, cameraTF)
    except Exception:
        display = crocoddyl.MeshcatDisplay(anymal)
    display.rate = -1
    display.freq = 1
    while True:
        display.displayFromSolver(boxddp)
        time.sleep(1.0)
        display.displayFromSolver(boxfddp)
        time.sleep(1.0)

# Plotting the entire motion
if WITHPLOT:
    plotSolution(boxfddp, figIndex=1, figTitle="Box-FDDP", show=False)
    plotSolution(boxddp, figIndex=3, figTitle="Box-DDP", show=False)

    log = boxfddp.getCallbacks()[1]
    crocoddyl.plotConvergence(
        log.costs,
        log.pregs,
        log.dregs,
        log.grads,
        log.stops,
        log.steps,
        show=False,
        figTitle="Box-FDDP",
        figIndex=5,
    )
    log = boxddp.getCallbacks()[1]
    crocoddyl.plotConvergence(
        log.costs,
        log.pregs,
        log.dregs,
        log.grads,
        log.stops,
        log.steps,
        figTitle="Box-DDP",
        figIndex=7,
    )



import os
import signal
import sys
import time

import example_robot_data
import numpy as np
import pinocchio

import crocoddyl
from crocoddyl.utils.biped import plotSolution

WITHDISPLAY = "display" in sys.argv or "CROCODDYL_DISPLAY" in os.environ
WITHPLOT = "plot" in sys.argv or "CROCODDYL_PLOT" in os.environ
signal.signal(signal.SIGINT, signal.SIG_DFL)

# Load robot
robot = example_robot_data.load("talos")
rmodel = robot.model
lims = rmodel.effortLimit
# lims[19:] *= 0.5  # reduced artificially the torque limits
rmodel.effortLimit = lims

# Create data structures
rdata = rmodel.createData()
state = crocoddyl.StateMultibody(rmodel)
actuation = crocoddyl.ActuationModelFloatingBase(state)

# Set integration time
DT = 5e-2
T = 40
target = np.array([0.4, 0, 1.2])

# Initialize reference state, target and reference CoM
rightFoot = "right_sole_link"
leftFoot = "left_sole_link"
endEffector = "gripper_left_joint"
endEffectorId = rmodel.getFrameId(endEffector)
rightFootId = rmodel.getFrameId(rightFoot)
leftFootId = rmodel.getFrameId(leftFoot)
q0 = rmodel.referenceConfigurations["half_sitting"]
x0 = np.concatenate([q0, np.zeros(rmodel.nv)])
pinocchio.forwardKinematics(rmodel, rdata, q0)
pinocchio.updateFramePlacements(rmodel, rdata)
rfPos0 = rdata.oMf[rightFootId].translation
lfPos0 = rdata.oMf[leftFootId].translation
refGripper = rdata.oMf[rmodel.getFrameId("gripper_left_joint")].translation
comRef = (rfPos0 + lfPos0) / 2
comRef[2] = pinocchio.centerOfMass(rmodel, rdata, q0)[2].item()

# Create two contact models used along the motion
contactModel1Foot = crocoddyl.ContactModelMultiple(state, actuation.nu)
contactModel2Feet = crocoddyl.ContactModelMultiple(state, actuation.nu)
supportContactModelLeft = crocoddyl.ContactModel6D(
    state,
    leftFootId,
    pinocchio.SE3.Identity(),
    pinocchio.LOCAL,
    actuation.nu,
    np.array([0, 40]),
)
supportContactModelRight = crocoddyl.ContactModel6D(
    state,
    rightFootId,
    pinocchio.SE3.Identity(),
    pinocchio.LOCAL,
    actuation.nu,
    np.array([0, 40]),
)
contactModel1Foot.addContact(rightFoot + "_contact", supportContactModelRight)
contactModel2Feet.addContact(leftFoot + "_contact", supportContactModelLeft)
contactModel2Feet.addContact(rightFoot + "_contact", supportContactModelRight)

# Cost for self-collision
maxfloat = sys.float_info.max
xlb = np.concatenate(
    [
        -maxfloat * np.ones(6),  # dimension of the SE(3) manifold
        rmodel.lowerPositionLimit[7:],
        -maxfloat * np.ones(state.nv),
    ]
)
xub = np.concatenate(
    [
        maxfloat * np.ones(6),  # dimension of the SE(3) manifold
        rmodel.upperPositionLimit[7:],
        maxfloat * np.ones(state.nv),
    ]
)
bounds = crocoddyl.ActivationBounds(xlb, xub, 1.0)
xLimitResidual = crocoddyl.ResidualModelState(state, x0, actuation.nu)
xLimitActivation = crocoddyl.ActivationModelQuadraticBarrier(bounds)
limitCost = crocoddyl.CostModelResidual(state, xLimitActivation, xLimitResidual)

# Cost for state and control
xResidual = crocoddyl.ResidualModelState(state, x0, actuation.nu)
xActivation = crocoddyl.ActivationModelWeightedQuad(
    np.array([0] * 3 + [10.0] * 3 + [0.01] * (state.nv - 6) + [10] * state.nv) ** 2
)
uResidual = crocoddyl.ResidualModelControl(state, actuation.nu)
xTActivation = crocoddyl.ActivationModelWeightedQuad(
    np.array([0] * 3 + [10.0] * 3 + [0.01] * (state.nv - 6) + [100] * state.nv) ** 2
)
xRegCost = crocoddyl.CostModelResidual(state, xActivation, xResidual)
uRegCost = crocoddyl.CostModelResidual(state, uResidual)
xRegTermCost = crocoddyl.CostModelResidual(state, xTActivation, xResidual)

# Cost for target reaching: hand and foot
handTrackingResidual = crocoddyl.ResidualModelFramePlacement(
    state, endEffectorId, pinocchio.SE3(np.eye(3), target), actuation.nu
)
handTrackingActivation = crocoddyl.ActivationModelWeightedQuad(
    np.array([1] * 3 + [0.0001] * 3) ** 2
)
handTrackingCost = crocoddyl.CostModelResidual(
    state, handTrackingActivation, handTrackingResidual
)

footTrackingResidual = crocoddyl.ResidualModelFramePlacement(
    state, leftFootId, pinocchio.SE3(np.eye(3), np.array([0.0, 0.4, 0.0])), actuation.nu
)
footTrackingActivation = crocoddyl.ActivationModelWeightedQuad(
    np.array([1, 1, 0.1] + [1.0] * 3) ** 2
)
footTrackingCost1 = crocoddyl.CostModelResidual(
    state, footTrackingActivation, footTrackingResidual
)
footTrackingResidual = crocoddyl.ResidualModelFramePlacement(
    state,
    leftFootId,
    pinocchio.SE3(np.eye(3), np.array([0.3, 0.15, 0.35])),
    actuation.nu,
)
footTrackingCost2 = crocoddyl.CostModelResidual(
    state, footTrackingActivation, footTrackingResidual
)

# Cost for CoM reference
comResidual = crocoddyl.ResidualModelCoMPosition(state, comRef, actuation.nu)
comTrack = crocoddyl.CostModelResidual(state, comResidual)

# Create cost model per each action model. We divide the motion in 3 phases plus its
# terminal model.
runningCostModel1 = crocoddyl.CostModelSum(state, actuation.nu)
runningCostModel2 = crocoddyl.CostModelSum(state, actuation.nu)
runningCostModel3 = crocoddyl.CostModelSum(state, actuation.nu)
terminalCostModel = crocoddyl.CostModelSum(state, actuation.nu)

# Then let's added the running and terminal cost functions
runningCostModel1.addCost("gripperPose", handTrackingCost, 1e2)
runningCostModel1.addCost("stateReg", xRegCost, 1e-3)
runningCostModel1.addCost("ctrlReg", uRegCost, 1e-4)
runningCostModel1.addCost("limitCost", limitCost, 1e3)

runningCostModel2.addCost("gripperPose", handTrackingCost, 1e2)
runningCostModel2.addCost("footPose", footTrackingCost1, 1e1)
runningCostModel2.addCost("stateReg", xRegCost, 1e-3)
runningCostModel2.addCost("ctrlReg", uRegCost, 1e-4)
runningCostModel2.addCost("limitCost", limitCost, 1e3)

runningCostModel3.addCost("gripperPose", handTrackingCost, 1e2)
runningCostModel3.addCost("footPose", footTrackingCost2, 1e1)
runningCostModel3.addCost("stateReg", xRegCost, 1e-3)
runningCostModel3.addCost("ctrlReg", uRegCost, 1e-4)
runningCostModel3.addCost("limitCost", limitCost, 1e3)

terminalCostModel.addCost("gripperPose", handTrackingCost, 1e2)
terminalCostModel.addCost("stateReg", xRegTermCost, 1e-3)
terminalCostModel.addCost("limitCost", limitCost, 1e3)

# Create the action model
dmodelRunning1 = crocoddyl.DifferentialActionModelContactFwdDynamics(
    state, actuation, contactModel2Feet, runningCostModel1
)
dmodelRunning2 = crocoddyl.DifferentialActionModelContactFwdDynamics(
    state, actuation, contactModel1Foot, runningCostModel2
)
dmodelRunning3 = crocoddyl.DifferentialActionModelContactFwdDynamics(
    state, actuation, contactModel1Foot, runningCostModel3
)
dmodelTerminal = crocoddyl.DifferentialActionModelContactFwdDynamics(
    state, actuation, contactModel1Foot, terminalCostModel
)

runningModel1 = crocoddyl.IntegratedActionModelEuler(dmodelRunning1, DT)
runningModel2 = crocoddyl.IntegratedActionModelEuler(dmodelRunning2, DT)
runningModel3 = crocoddyl.IntegratedActionModelEuler(dmodelRunning3, DT)
terminalModel = crocoddyl.IntegratedActionModelEuler(dmodelTerminal, 0)

# Problem definition
x0 = np.concatenate([q0, pinocchio.utils.zero(state.nv)])
problem = crocoddyl.ShootingProblem(
    x0, [runningModel1] * T + [runningModel2] * T + [runningModel3] * T, terminalModel
)

# Creating the DDP solver for this OC problem, defining a logger
solver = crocoddyl.SolverBoxFDDP(problem)
if WITHPLOT:
    solver.setCallbacks(
        [
            crocoddyl.CallbackVerbose(),
            crocoddyl.CallbackLogger(),
        ]
    )
else:
    solver.setCallbacks([crocoddyl.CallbackVerbose()])

# Solving it with the DDP algorithm
xs = [x0] * (solver.problem.T + 1)
us = solver.problem.quasiStatic([x0] * solver.problem.T)
solver.th_stop = 1e-7
solver.solve(xs, us, 500, False, 1e-9)

# Visualizing the solution in gepetto-viewer
display = None
if WITHDISPLAY:
    if display is None:
        try:
            import gepetto

            gepetto.corbaserver.Client()
            display = crocoddyl.GepettoDisplay(robot)
            display.robot.viewer.gui.addSphere(
                "world/point", 0.05, [1.0, 0.0, 0.0, 1.0]
            )  # radius = .1, RGBA=1001
            display.robot.viewer.gui.applyConfiguration(
                "world/point", [*target.tolist(), 0.0, 0.0, 0.0, 1.0]
            )  # xyz+quaternion
        except Exception:
            display = crocoddyl.MeshcatDisplay(robot)
    display.rate = -1
    display.freq = 1
    while True:
        display.displayFromSolver(solver)
        time.sleep(1.0)

# Get final state and end effector position
xT = solver.xs[-1]
pinocchio.forwardKinematics(rmodel, rdata, xT[: state.nq])
pinocchio.updateFramePlacements(rmodel, rdata)
com = pinocchio.centerOfMass(rmodel, rdata, xT[: state.nq])
finalPosEff = np.array(
    rdata.oMf[rmodel.getFrameId("gripper_left_joint")].translation.T.flat
)

print("Finally reached = ({:.3f}, {:.3f}, {:.3f})".format(*finalPosEff))
print(f"Distance between hand and target = {np.linalg.norm(finalPosEff - target):.3E}")
print(f"Distance to default state = {np.linalg.norm(x0 - np.array(xT.flat)):.3E}")
print(f"XY distance to CoM reference = {np.linalg.norm(com[:2] - comRef[:2]):.3E}")

# Plotting the entire motion
if WITHPLOT:
    log = solver.getCallbacks()[1]
    plotSolution(solver, bounds=False, figIndex=1, show=False)

    crocoddyl.plotConvergence(
        log.costs, log.pregs, log.dregs, log.grads, log.stops, log.steps, figIndex=3
    )



import numpy as np
import pinocchio

import crocoddyl


class SimpleBipedGaitProblem:
    """Build simple bipedal locomotion problems.

    This class aims to build simple locomotion problems used in the examples of
    Crocoddyl.
    The scope of this class is purely for academic reasons, and it does not aim to be
    used in any robotics application.
    We also do not consider it as part of the API, so changes in this class will not
    pass through a strict process of deprecation.
    Thus, we advice any user to DO NOT develop their application based on this class.
    """

    def __init__(
        self,
        rmodel,
        rightFoot,
        leftFoot,
        integrator="euler",
        control="zero",
        fwddyn=True,
    ):
        """Construct biped-gait problem.

        :param rmodel: robot model
        :param rightFoot: name of the right foot
        :param leftFoot: name of the left foot
        :param integrator: type of the integrator
            (options are: 'euler', and 'rk4')
        :param control: type of control parametrization
            (options are: 'zero', 'one', and 'rk4')
        """
        self.rmodel = rmodel
        self.rdata = rmodel.createData()
        self.state = crocoddyl.StateMultibody(self.rmodel)
        self.actuation = crocoddyl.ActuationModelFloatingBase(self.state)
        # Getting the frame id for all the legs
        self.rfId = self.rmodel.getFrameId(rightFoot)
        self.lfId = self.rmodel.getFrameId(leftFoot)
        self._integrator = integrator
        self._control = control
        self._fwddyn = fwddyn
        # Defining default state
        q0 = self.rmodel.referenceConfigurations["half_sitting"]
        self.rmodel.defaultState = np.concatenate([q0, np.zeros(self.rmodel.nv)])
        self.firstStep = True
        # Defining the friction coefficient and normal
        self.mu = 0.7
        self.Rsurf = np.eye(3)

    def createWalkingProblem(
        self, x0, stepLength, stepHeight, timeStep, stepKnots, supportKnots
    ):
        """Create a shooting problem for a simple walking gait.

        :param x0: initial state
        :param stepLength: step length
        :param stepHeight: step height
        :param timeStep: step time for each knot
        :param stepKnots: number of knots for step phases
        :param supportKnots: number of knots for double support phases
        :return shooting problem
        """
        # Compute the current foot positions
        q0 = x0[: self.state.nq]
        pinocchio.forwardKinematics(self.rmodel, self.rdata, q0)
        pinocchio.updateFramePlacements(self.rmodel, self.rdata)
        rfPos0 = self.rdata.oMf[self.rfId].translation
        lfPos0 = self.rdata.oMf[self.lfId].translation
        comRef = (rfPos0 + lfPos0) / 2
        comRef[2] = pinocchio.centerOfMass(self.rmodel, self.rdata, q0)[2]

        # Defining the action models along the time instances
        loco3dModel = []
        doubleSupport = [
            self.createSwingFootModel(timeStep, [self.rfId, self.lfId])
            for k in range(supportKnots)
        ]

        # Creating the action models for three steps
        if self.firstStep is True:
            rStep = self.createFootstepModels(
                comRef,
                [rfPos0],
                0.5 * stepLength,
                stepHeight,
                timeStep,
                stepKnots,
                [self.lfId],
                [self.rfId],
            )
            self.firstStep = False
        else:
            rStep = self.createFootstepModels(
                comRef,
                [rfPos0],
                stepLength,
                stepHeight,
                timeStep,
                stepKnots,
                [self.lfId],
                [self.rfId],
            )
        lStep = self.createFootstepModels(
            comRef,
            [lfPos0],
            stepLength,
            stepHeight,
            timeStep,
            stepKnots,
            [self.rfId],
            [self.lfId],
        )

        # We defined the problem as:
        loco3dModel += doubleSupport + rStep
        loco3dModel += doubleSupport + lStep

        problem = crocoddyl.ShootingProblem(x0, loco3dModel[:-1], loco3dModel[-1])
        return problem

    def createJumpingProblem(
        self,
        x0,
        jumpHeight,
        jumpLength,
        timeStep,
        groundKnots,
        flyingKnots,
        final=False,
    ):
        q0 = x0[: self.rmodel.nq]
        pinocchio.forwardKinematics(self.rmodel, self.rdata, q0)
        pinocchio.updateFramePlacements(self.rmodel, self.rdata)
        rfFootPos0 = self.rdata.oMf[self.rfId].translation
        lfFootPos0 = self.rdata.oMf[self.lfId].translation
        df = jumpLength[2] - rfFootPos0[2]
        rfFootPos0[2] = 0.0
        lfFootPos0[2] = 0.0
        comRef = (rfFootPos0 + lfFootPos0) / 2
        comRef[2] = pinocchio.centerOfMass(self.rmodel, self.rdata, q0)[2]

        self.rWeight = 1e1
        loco3dModel = []
        takeOff = [
            self.createSwingFootModel(
                timeStep,
                [self.lfId, self.rfId],
            )
            for k in range(groundKnots)
        ]
        flyingUpPhase = [
            self.createSwingFootModel(
                timeStep,
                [],
                np.array([jumpLength[0], jumpLength[1], jumpLength[2] + jumpHeight])
                * (k + 1)
                / flyingKnots
                + comRef,
            )
            for k in range(flyingKnots)
        ]
        flyingDownPhase = []
        for k in range(flyingKnots):
            flyingDownPhase += [self.createSwingFootModel(timeStep, [])]

        f0 = jumpLength
        footTask = [
            [self.lfId, pinocchio.SE3(np.eye(3), self.lfId + f0)],
            [self.rfId, pinocchio.SE3(np.eye(3), self.rfId + f0)],
        ]
        landingPhase = [
            self.createFootSwitchModel([self.lfId, self.rfId], footTask, False)
        ]
        f0[2] = df
        if final is True:
            self.rWeight = 1e4
        landed = [
            self.createSwingFootModel(
                timeStep, [self.lfId, self.rfId], comTask=comRef + f0
            )
            for k in range(groundKnots)
        ]
        loco3dModel += takeOff
        loco3dModel += flyingUpPhase
        loco3dModel += flyingDownPhase
        loco3dModel += landingPhase
        loco3dModel += landed

        # Rescaling the terminal weights
        costs = loco3dModel[-1].differential.costs.costs.todict()
        for c in costs.values():
            c.weight *= timeStep

        problem = crocoddyl.ShootingProblem(x0, loco3dModel[:-1], loco3dModel[-1])
        return problem

    def createFootstepModels(
        self,
        comPos0,
        feetPos0,
        stepLength,
        stepHeight,
        timeStep,
        numKnots,
        supportFootIds,
        swingFootIds,
    ):
        """Action models for a footstep phase.

        :param comPos0, initial CoM position
        :param feetPos0: initial position of the swinging feet
        :param stepLength: step length
        :param stepHeight: step height
        :param timeStep: time step
        :param numKnots: number of knots for the footstep phase
        :param supportFootIds: Ids of the supporting feet
        :param swingFootIds: Ids of the swinging foot
        :return footstep action models
        """
        numLegs = len(supportFootIds) + len(swingFootIds)
        comPercentage = float(len(swingFootIds)) / numLegs

        # Action models for the foot swing
        footSwingModel = []
        for k in range(numKnots):
            swingFootTask = []
            for i, p in zip(swingFootIds, feetPos0):
                # Defining a foot swing task given the step length. The swing task
                # is decomposed on two phases: swing-up and swing-down. We decide
                # deliveratively to allocated the same number of nodes (i.e. phKnots)
                # in each phase. With this, we define a proper z-component for the
                # swing-leg motion.
                phKnots = numKnots / 2
                if k < phKnots:
                    dp = np.array(
                        [stepLength * (k + 1) / numKnots, 0.0, stepHeight * k / phKnots]
                    )
                elif k == phKnots:
                    dp = np.array([stepLength * (k + 1) / numKnots, 0.0, stepHeight])
                else:
                    dp = np.array(
                        [
                            stepLength * (k + 1) / numKnots,
                            0.0,
                            stepHeight * (1 - float(k - phKnots) / phKnots),
                        ]
                    )
                tref = p + dp

                swingFootTask += [[i, pinocchio.SE3(np.eye(3), tref)]]

            comTask = (
                np.array([stepLength * (k + 1) / numKnots, 0.0, 0.0]) * comPercentage
                + comPos0
            )
            footSwingModel += [
                self.createSwingFootModel(
                    timeStep,
                    supportFootIds,
                    comTask=comTask,
                    swingFootTask=swingFootTask,
                )
            ]

        # Action model for the foot switch
        footSwitchModel = self.createFootSwitchModel(supportFootIds, swingFootTask)

        # Updating the current foot position for next step
        comPos0 += [stepLength * comPercentage, 0.0, 0.0]
        for p in feetPos0:
            p += [stepLength, 0.0, 0.0]
        return [*footSwingModel, footSwitchModel]

    def createSwingFootModel(
        self, timeStep, supportFootIds, comTask=None, swingFootTask=None
    ):
        """Action model for a swing foot phase.

        :param timeStep: step duration of the action model
        :param supportFootIds: Ids of the constrained feet
        :param comTask: CoM task
        :param swingFootTask: swinging foot task
        :return action model for a swing foot phase
        """
        # Creating a 6D multi-contact model, and then including the supporting
        # foot
        if self._fwddyn:
            nu = self.actuation.nu
        else:
            nu = self.state.nv + 6 * len(supportFootIds)
        contactModel = crocoddyl.ContactModelMultiple(self.state, nu)
        for i in supportFootIds:
            supportContactModel = crocoddyl.ContactModel6D(
                self.state,
                i,
                pinocchio.SE3.Identity(),
                pinocchio.LOCAL_WORLD_ALIGNED,
                nu,
                np.array([0.0, 50.0]),
            )
            contactModel.addContact(
                self.rmodel.frames[i].name + "_contact", supportContactModel
            )

        # Creating the cost model for a contact phase
        costModel = crocoddyl.CostModelSum(self.state, nu)
        if isinstance(comTask, np.ndarray):
            comResidual = crocoddyl.ResidualModelCoMPosition(self.state, comTask, nu)
            comTrack = crocoddyl.CostModelResidual(self.state, comResidual)
            costModel.addCost("comTrack", comTrack, 1e6)
        for i in supportFootIds:
            cone = crocoddyl.WrenchCone(self.Rsurf, self.mu, np.array([0.1, 0.05]))
            wrenchResidual = crocoddyl.ResidualModelContactWrenchCone(
                self.state, i, cone, nu, self._fwddyn
            )
            wrenchActivation = crocoddyl.ActivationModelQuadraticBarrier(
                crocoddyl.ActivationBounds(cone.lb, cone.ub)
            )
            wrenchCone = crocoddyl.CostModelResidual(
                self.state, wrenchActivation, wrenchResidual
            )
            costModel.addCost(
                self.rmodel.frames[i].name + "_wrenchCone", wrenchCone, 1e1
            )

        if swingFootTask is not None:
            for i in swingFootTask:
                framePlacementResidual = crocoddyl.ResidualModelFramePlacement(
                    self.state, i[0], i[1], nu
                )
                footTrack = crocoddyl.CostModelResidual(
                    self.state, framePlacementResidual
                )
                costModel.addCost(
                    self.rmodel.frames[i[0]].name + "_footTrack", footTrack, 1e6
                )

        stateWeights = np.array(
            [0] * 3 + [500.0] * 3 + [0.01] * (self.state.nv - 6) + [10] * self.state.nv
        )
        stateResidual = crocoddyl.ResidualModelState(
            self.state, self.rmodel.defaultState, nu
        )
        stateActivation = crocoddyl.ActivationModelWeightedQuad(stateWeights**2)
        stateReg = crocoddyl.CostModelResidual(
            self.state, stateActivation, stateResidual
        )
        if self._fwddyn:
            ctrlResidual = crocoddyl.ResidualModelControl(self.state, nu)
        else:
            ctrlResidual = crocoddyl.ResidualModelJointEffort(
                self.state, self.actuation, nu
            )
        ctrlReg = crocoddyl.CostModelResidual(self.state, ctrlResidual)
        costModel.addCost("stateReg", stateReg, 1e1)
        costModel.addCost("ctrlReg", ctrlReg, 1e-1)

        # Creating the action model for the KKT dynamics with simpletic Euler
        # integration scheme
        if self._fwddyn:
            dmodel = crocoddyl.DifferentialActionModelContactFwdDynamics(
                self.state, self.actuation, contactModel, costModel, 0.0, True
            )
        else:
            dmodel = crocoddyl.DifferentialActionModelContactInvDynamics(
                self.state, self.actuation, contactModel, costModel
            )
        if self._control == "one":
            control = crocoddyl.ControlParametrizationModelPolyOne(nu)
        elif self._control == "rk4":
            control = crocoddyl.ControlParametrizationModelPolyTwoRK(
                nu, crocoddyl.RKType.four
            )
        elif self._control == "rk3":
            control = crocoddyl.ControlParametrizationModelPolyTwoRK(
                nu, crocoddyl.RKType.three
            )
        else:
            control = crocoddyl.ControlParametrizationModelPolyZero(nu)
        if self._integrator == "euler":
            model = crocoddyl.IntegratedActionModelEuler(dmodel, control, timeStep)
        elif self._integrator == "rk4":
            model = crocoddyl.IntegratedActionModelRK(
                dmodel, control, crocoddyl.RKType.four, timeStep
            )
        elif self._integrator == "rk3":
            model = crocoddyl.IntegratedActionModelRK(
                dmodel, control, crocoddyl.RKType.three, timeStep
            )
        elif self._integrator == "rk2":
            model = crocoddyl.IntegratedActionModelRK(
                dmodel, control, crocoddyl.RKType.two, timeStep
            )
        else:
            model = crocoddyl.IntegratedActionModelEuler(dmodel, control, timeStep)
        return model

    def createFootSwitchModel(self, supportFootIds, swingFootTask, pseudoImpulse=False):
        """Action model for a foot switch phase.

        :param supportFootIds: Ids of the constrained feet
        :param swingFootTask: swinging foot task
        :param pseudoImpulse: true for pseudo-impulse models, otherwise it uses the
            impulse model
        :return action model for a foot switch phase
        """
        if pseudoImpulse:
            return self.createPseudoImpulseModel(supportFootIds, swingFootTask)
        else:
            return self.createImpulseModel(supportFootIds, swingFootTask)

    def createPseudoImpulseModel(self, supportFootIds, swingFootTask):
        """Action model for pseudo-impulse models.

        A pseudo-impulse model consists of adding high-penalty cost for the contact
        velocities.
        :param swingFootTask: swinging foot task
        :return pseudo-impulse differential action model
        """

        # Creating a 6D multi-contact model, and then including the supporting
        # foot
        if self._fwddyn:
            nu = self.actuation.nu
        else:
            nu = self.state.nv + 6 * len(supportFootIds)
        contactModel = crocoddyl.ContactModelMultiple(self.state, nu)
        for i in supportFootIds:
            supportContactModel = crocoddyl.ContactModel6D(
                self.state,
                i,
                pinocchio.SE3.Identity(),
                pinocchio.LOCAL_WORLD_ALIGNED,
                nu,
                np.array([0.0, 50.0]),
            )
            contactModel.addContact(
                self.rmodel.frames[i].name + "_contact", supportContactModel
            )

        # Creating the cost model for a contact phase
        costModel = crocoddyl.CostModelSum(self.state, nu)
        for i in supportFootIds:
            cone = crocoddyl.WrenchCone(self.Rsurf, self.mu, np.array([0.1, 0.05]))
            wrenchResidual = crocoddyl.ResidualModelContactWrenchCone(
                self.state, i, cone, nu, self._fwddyn
            )
            wrenchActivation = crocoddyl.ActivationModelQuadraticBarrier(
                crocoddyl.ActivationBounds(cone.lb, cone.ub)
            )
            wrenchCone = crocoddyl.CostModelResidual(
                self.state, wrenchActivation, wrenchResidual
            )
            costModel.addCost(
                self.rmodel.frames[i].name + "_wrenchCone", wrenchCone, 1e1
            )

        if swingFootTask is not None:
            for i in swingFootTask:
                framePlacementResidual = crocoddyl.ResidualModelFramePlacement(
                    self.state, i[0], i[1], nu
                )
                frameVelocityResidual = crocoddyl.ResidualModelFrameVelocity(
                    self.state,
                    i[0],
                    pinocchio.Motion.Zero(),
                    pinocchio.LOCAL_WORLD_ALIGNED,
                    nu,
                )
                footTrack = crocoddyl.CostModelResidual(
                    self.state, framePlacementResidual
                )
                impulseFootVelCost = crocoddyl.CostModelResidual(
                    self.state, frameVelocityResidual
                )
                costModel.addCost(
                    self.rmodel.frames[i[0]].name + "_footTrack", footTrack, 1e8
                )
                costModel.addCost(
                    self.rmodel.frames[i[0]].name + "_impulseVel",
                    impulseFootVelCost,
                    1e6,
                )

        stateWeights = np.array(
            [0.0] * 3
            + [500.0] * 3
            + [0.01] * (self.state.nv - 6)
            + [10] * self.state.nv
        )
        stateResidual = crocoddyl.ResidualModelState(
            self.state, self.rmodel.defaultState, nu
        )
        stateActivation = crocoddyl.ActivationModelWeightedQuad(stateWeights**2)
        stateReg = crocoddyl.CostModelResidual(
            self.state, stateActivation, stateResidual
        )
        if self._fwddyn:
            ctrlResidual = crocoddyl.ResidualModelControl(self.state, nu)
            ctrlReg = crocoddyl.CostModelResidual(self.state, ctrlResidual)
        else:
            ctrlResidual = crocoddyl.ResidualModelJointEffort(
                self.state, self.actuation, nu
            )
            ctrlReg = crocoddyl.CostModelResidual(self.state, ctrlResidual)
        costModel.addCost("stateReg", stateReg, 1e1)
        costModel.addCost("ctrlReg", ctrlReg, 1e-3)

        # Creating the action model for the KKT dynamics with simpletic Euler
        # integration scheme
        if self._fwddyn:
            dmodel = crocoddyl.DifferentialActionModelContactFwdDynamics(
                self.state, self.actuation, contactModel, costModel, 0.0, True
            )
        else:
            dmodel = crocoddyl.DifferentialActionModelContactInvDynamics(
                self.state, self.actuation, contactModel, costModel
            )
        if self._integrator == "euler":
            model = crocoddyl.IntegratedActionModelEuler(dmodel, 0.0)
        elif self._integrator == "rk4":
            model = crocoddyl.IntegratedActionModelRK(
                dmodel, crocoddyl.RKType.four, 0.0
            )
        elif self._integrator == "rk3":
            model = crocoddyl.IntegratedActionModelRK(
                dmodel, crocoddyl.RKType.three, 0.0
            )
        elif self._integrator == "rk2":
            model = crocoddyl.IntegratedActionModelRK(dmodel, crocoddyl.RKType.two, 0.0)
        else:
            model = crocoddyl.IntegratedActionModelEuler(dmodel, 0.0)
        return model

    def createImpulseModel(self, supportFootIds, swingFootTask):
        """Action model for impulse models.

        An impulse model consists of describing the impulse dynamics against a set of
        contacts.
        :param supportFootIds: Ids of the constrained feet
        :param swingFootTask: swinging foot task
        :return impulse action model
        """
        # Creating a 6D multi-contact model, and then including the supporting foot
        impulseModel = crocoddyl.ImpulseModelMultiple(self.state)
        for i in supportFootIds:
            supportContactModel = crocoddyl.ImpulseModel6D(
                self.state, i, pinocchio.LOCAL_WORLD_ALIGNED
            )
            impulseModel.addImpulse(
                self.rmodel.frames[i].name + "_impulse", supportContactModel
            )

        # Creating the cost model for a contact phase
        costModel = crocoddyl.CostModelSum(self.state, 0)
        if swingFootTask is not None:
            for i in swingFootTask:
                frameTranslationResidual = crocoddyl.ResidualModelFrameTranslation(
                    self.state, i[0], i[1].translation, 0
                )
                footTrack = crocoddyl.CostModelResidual(
                    self.state, frameTranslationResidual
                )
                costModel.addCost(
                    self.rmodel.frames[i[0]].name + "_footTrack", footTrack, 1e8
                )

        stateWeights = np.array(
            [1.0] * 6 + [0.1] * (self.rmodel.nv - 6) + [10] * self.rmodel.nv
        )
        stateResidual = crocoddyl.ResidualModelState(
            self.state, self.rmodel.defaultState, 0
        )
        stateActivation = crocoddyl.ActivationModelWeightedQuad(stateWeights**2)
        stateReg = crocoddyl.CostModelResidual(
            self.state, stateActivation, stateResidual
        )
        costModel.addCost("stateReg", stateReg, 1e1)

        # Creating the action model for the KKT dynamics with simpletic Euler
        # integration scheme
        model = crocoddyl.ActionModelImpulseFwdDynamics(
            self.state, impulseModel, costModel
        )
        return model


def plotSolution(solver, bounds=True, figIndex=1, figTitle="", show=True):
    import matplotlib.pyplot as plt

    xs, us, cs = [], [], []
    if bounds:
        us_lb, us_ub = [], []
        xs_lb, xs_ub = [], []

    def updateTrajectories(solver):
        xs.extend(solver.xs[:-1])
        for m, d in zip(solver.problem.runningModels, solver.problem.runningDatas):
            if hasattr(m, "differential"):
                cs.append(d.differential.multibody.pinocchio.com[0])
                us.append(d.differential.multibody.joint.tau)
                if bounds and isinstance(
                    m.differential, crocoddyl.DifferentialActionModelContactFwdDynamics
                ):
                    us_lb.extend([m.u_lb])
                    us_ub.extend([m.u_ub])
            else:
                cs.append(d.multibody.pinocchio.com[0])
                us.append(np.zeros(nu))
                if bounds:
                    us_lb.append(np.nan * np.ones(nu))
                    us_ub.append(np.nan * np.ones(nu))
            if bounds:
                xs_lb.extend([m.state.lb])
                xs_ub.extend([m.state.ub])

    if isinstance(solver, list):
        for s in solver:
            rmodel = solver[0].problem.runningModels[0].state.pinocchio
            nq, nv, nu = (
                rmodel.nq,
                rmodel.nv,
                solver[0].problem.runningModels[0].differential.actuation.nu,
            )
            updateTrajectories(s)
    else:
        rmodel = solver.problem.runningModels[0].state.pinocchio
        nq, nv, nu = (
            rmodel.nq,
            rmodel.nv,
            solver.problem.runningModels[0].differential.actuation.nu,
        )
        updateTrajectories(solver)

    # Getting the state and control trajectories
    nx = nq + nv
    X = [0.0] * nx
    U = [0.0] * nu
    if bounds:
        U_LB = [0.0] * nu
        U_UB = [0.0] * nu
        X_LB = [0.0] * nx
        X_UB = [0.0] * nx
    for i in range(nx):
        X[i] = [x[i] for x in xs]
        if bounds:
            X_LB[i] = [x[i] for x in xs_lb]
            X_UB[i] = [x[i] for x in xs_ub]
    for i in range(nu):
        U[i] = [u[i] for u in us]
        if bounds:
            U_LB[i] = [u[i] for u in us_lb]
            U_UB[i] = [u[i] for u in us_ub]

    # Plotting the joint positions, velocities and torques
    plt.figure(figIndex)
    plt.suptitle(figTitle)
    legJointNames = ["1", "2", "3", "4", "5", "6"]
    # left foot
    plt.subplot(2, 3, 1)
    plt.title("joint position [rad]")
    [plt.plot(X[k], label=legJointNames[i]) for i, k in enumerate(range(7, 13))]
    if bounds:
        [plt.plot(X_LB[k], "--r") for i, k in enumerate(range(7, 13))]
        [plt.plot(X_UB[k], "--r") for i, k in enumerate(range(7, 13))]
    plt.ylabel("LF")
    plt.legend()
    plt.subplot(2, 3, 2)
    plt.title("joint velocity [rad/s]")
    [
        plt.plot(X[k], label=legJointNames[i])
        for i, k in enumerate(range(nq + 6, nq + 12))
    ]
    if bounds:
        [plt.plot(X_LB[k], "--r") for i, k in enumerate(range(nq + 6, nq + 12))]
        [plt.plot(X_UB[k], "--r") for i, k in enumerate(range(nq + 6, nq + 12))]
    plt.ylabel("LF")
    plt.legend()
    plt.subplot(2, 3, 3)
    plt.title("joint torque [Nm]")
    [plt.plot(U[k], label=legJointNames[i]) for i, k in enumerate(range(0, 6))]
    if bounds:
        [plt.plot(U_LB[k], "--r") for i, k in enumerate(range(0, 6))]
        [plt.plot(U_UB[k], "--r") for i, k in enumerate(range(0, 6))]
    plt.ylabel("LF")
    plt.legend()

    # right foot
    plt.subplot(2, 3, 4)
    [plt.plot(X[k], label=legJointNames[i]) for i, k in enumerate(range(13, 19))]
    if bounds:
        [plt.plot(X_LB[k], "--r") for i, k in enumerate(range(13, 19))]
        [plt.plot(X_UB[k], "--r") for i, k in enumerate(range(13, 19))]
    plt.ylabel("RF")
    plt.xlabel("knots")
    plt.legend()
    plt.subplot(2, 3, 5)
    [
        plt.plot(X[k], label=legJointNames[i])
        for i, k in enumerate(range(nq + 12, nq + 18))
    ]
    if bounds:
        [plt.plot(X_LB[k], "--r") for i, k in enumerate(range(nq + 12, nq + 18))]
        [plt.plot(X_UB[k], "--r") for i, k in enumerate(range(nq + 12, nq + 18))]
    plt.ylabel("RF")
    plt.xlabel("knots")
    plt.legend()
    plt.subplot(2, 3, 6)
    [plt.plot(U[k], label=legJointNames[i]) for i, k in enumerate(range(6, 12))]
    if bounds:
        [plt.plot(U_LB[k], "--r") for i, k in enumerate(range(6, 12))]
        [plt.plot(U_UB[k], "--r") for i, k in enumerate(range(6, 12))]
    plt.ylabel("RF")
    plt.xlabel("knots")
    plt.legend()

    plt.figure(figIndex + 1)
    rdata = rmodel.createData()
    Cx = []
    Cy = []
    for x in xs:
        q = x[: rmodel.nq]
        c = pinocchio.centerOfMass(rmodel, rdata, q)
        Cx.append(c[0])
        Cy.append(c[1])
    plt.plot(Cx, Cy)
    plt.title("CoM position")
    plt.xlabel("x [m]")
    plt.ylabel("y [m]")
    plt.grid(True)
    if show:
        plt.show()



import sys
import unittest
from random import randint

import example_robot_data
import numpy as np
from factory import StateMultibodyDerived, StateVectorDerived

import crocoddyl


class StateAbstractTestCase(unittest.TestCase):
    NX = None
    NDX = None
    STATE = None
    STATE_DER = None

    def test_state_dimension(self):
        # Checking the state dimensions
        self.assertEqual(self.STATE.nx, self.STATE_DER.nx, "Wrong nx value.")
        self.assertEqual(self.STATE.ndx, self.STATE_DER.ndx, "Wrong ndx value.")
        self.assertEqual(self.STATE.nq, self.STATE_DER.nq, "Wrong nq value.")
        self.assertEqual(self.STATE.nv, self.STATE_DER.nv, "Wrong nv value.")

        # Checking the dimension of zero and random states
        self.assertEqual(
            self.STATE.zero().shape, (self.NX,), "Wrong dimension of zero state."
        )
        self.assertEqual(
            self.STATE.rand().shape, (self.NX,), "Wrong dimension of random state."
        )
        self.assertEqual(
            self.STATE.lb.shape, (self.NX,), "Wrong dimension of lower bound."
        )
        self.assertEqual(
            self.STATE.ub.shape, (self.NX,), "Wrong dimension of lower bound."
        )

    def test_python_derived_diff(self):
        x0 = self.STATE.rand()
        x1 = self.STATE.rand()

        # Checking that both diff functions agree
        self.assertTrue(
            np.allclose(
                self.STATE.diff(x0, x1), self.STATE_DER.diff(x0, x1), atol=1e-9
            ),
            "state.diff() function doesn't agree with Python bindings.",
        )

    def test_python_derived_integrate(self):
        x = self.STATE.rand()
        dx = self.STATE.rand()[: self.STATE.ndx]

        # Checking that both integrate functions agree
        self.assertTrue(
            np.allclose(
                self.STATE.integrate(x, dx), self.STATE_DER.integrate(x, dx), atol=1e-9
            ),
            "state.integrate() function doesn't agree with Python bindings.",
        )

    def test_python_derived_Jdiff(self):
        x0 = self.STATE.rand()
        x1 = self.STATE.rand()

        # Checking that both Jdiff functions agree
        J1, J2 = self.STATE.Jdiff(x0, x1)
        J1d, J2d = self.STATE_DER.Jdiff(x0, x1)
        if self.STATE.__class__ == crocoddyl.libcrocoddyl_pywrap.StateMultibody:
            nv = self.STATE.nv
            self.assertTrue(
                np.allclose(J1[nv:, nv:], J1d[nv:, nv:], atol=1e-9),
                "state.Jdiff()[0] function doesn't agree with Python bindings.",
            )
            self.assertTrue(
                np.allclose(J2[nv:, nv:], J2d[nv:, nv:], atol=1e-9),
                "state.Jdiff()[1] function doesn't agree with Python bindings.",
            )
        else:
            self.assertTrue(
                np.allclose(J1, J1d, atol=1e-9),
                "state.Jdiff()[0] function doesn't agree with Python bindings.",
            )
            self.assertTrue(
                np.allclose(J2, J2d, atol=1e-9),
                "state.Jdiff()[1] function doesn't agree with Python bindings.",
            )

    def test_python_derived_Jintegrate(self):
        x = self.STATE.rand()
        dx = self.STATE.rand()[: self.STATE.ndx]

        # Checking that both Jintegrate functions agree
        J1, J2 = self.STATE.Jintegrate(x, dx)
        J1d, J2d = self.STATE_DER.Jintegrate(x, dx)
        if self.STATE.__class__ == crocoddyl.libcrocoddyl_pywrap.StateMultibody:
            nv = self.STATE.nv
            self.assertTrue(
                np.allclose(J1[nv:, nv:], J1d[nv:, nv:], atol=1e-9),
                "state.Jintegrate()[0] function doesn't agree with Python bindings.",
            )
            self.assertTrue(
                np.allclose(J2[nv:, nv:], J2d[nv:, nv:], atol=1e-9),
                "state.Jintegrate()[1] function doesn't agree with Python bindings.",
            )
        else:
            self.assertTrue(
                np.allclose(J1, J1d, atol=1e-9),
                "state.Jintegrate()[0] function doesn't agree with Python bindings.",
            )
            self.assertTrue(
                np.allclose(J2, J2d, atol=1e-9),
                "state.Jintegrate()[1] function doesn't agree with Python bindings.",
            )


class StateVectorTest(StateAbstractTestCase):
    NX = randint(1, 101)
    NDX = StateAbstractTestCase.NX
    STATE = crocoddyl.StateVector(NX)
    STATE_DER = StateVectorDerived(NX)


class StateMultibodyTalosArmTest(StateAbstractTestCase):
    MODEL = example_robot_data.load("talos_arm").model
    NX = MODEL.nq + MODEL.nv
    NDX = 2 * MODEL.nv
    STATE = crocoddyl.StateMultibody(MODEL)
    STATE_DER = StateMultibodyDerived(MODEL)


class StateMultibodyHyQTest(StateAbstractTestCase):
    MODEL = example_robot_data.load("hyq").model
    NX = MODEL.nq + MODEL.nv
    NDX = 2 * MODEL.nv
    STATE = crocoddyl.StateMultibody(MODEL)
    STATE_DER = StateMultibodyDerived(MODEL)


class StateMultibodyTalosTest(StateAbstractTestCase):
    MODEL = example_robot_data.load("talos").model
    NX = MODEL.nq + MODEL.nv
    NDX = 2 * MODEL.nv
    STATE = crocoddyl.StateMultibody(MODEL)
    STATE_DER = StateMultibodyDerived(MODEL)


if __name__ == "__main__":
    # test to be run
    test_classes_to_run = [
        StateVectorTest,
        StateMultibodyTalosArmTest,
        StateMultibodyHyQTest,
        StateMultibodyTalosTest,
    ]
    loader = unittest.TestLoader()
    suites_list = []
    for test_class in test_classes_to_run:
        suite = loader.loadTestsFromTestCase(test_class)
        suites_list.append(suite)
    big_suite = unittest.TestSuite(suites_list)
    runner = unittest.TextTestRunner()
    results = runner.run(big_suite)
    sys.exit(not results.wasSuccessful())



import copy
import sys
import unittest

import numpy as np
import pinocchio

import crocoddyl


class CopyModelTestCase(unittest.TestCase):
    MODEL = list()
    DATA = False
    COLLECTOR = list()

    def test_copy(self):
        Mcopy = copy.copy(self.MODEL)
        self.assertFalse(id(self.MODEL) == id(Mcopy))
        if self.DATA:
            D = []
            for i, m in enumerate(self.MODEL):
                if not self.COLLECTOR:
                    D.append(m.createData())
                else:
                    D.append(m.createData(self.COLLECTOR[i]))
            Dcopy = copy.copy(D)
            self.assertFalse(id(D) == id(Dcopy))
        for i, m in enumerate(self.MODEL):
            self.assertTrue(id(self.MODEL[i]) == id(Mcopy[i]))
            if self.DATA:
                self.assertTrue(id(D[i]) == id(Dcopy[i]))

    def test_deepcopy(self):
        Mcopy = copy.deepcopy(self.MODEL)
        self.assertFalse(id(self.MODEL) == id(Mcopy))
        if self.DATA:
            D = []
            for i, m in enumerate(self.MODEL):
                if not self.COLLECTOR:
                    D.append(m.createData())
                else:
                    D.append(m.createData(self.COLLECTOR[i]))
            Dcopy = copy.deepcopy(D)
            self.assertFalse(id(D) == id(Dcopy))
        for i, m in enumerate(self.MODEL):
            self.assertFalse(id(self.MODEL[i]) == id(Mcopy[i]))
            if self.DATA:
                self.assertFalse(id(D[i]) == id(Dcopy[i]))


class ActionsTest(CopyModelTestCase):
    MODEL = list()
    DATA = True
    # core actions
    MODEL.append(crocoddyl.ActionModelUnicycle())
    MODEL.append(crocoddyl.ActionModelLQR(2, 2))
    MODEL.append(crocoddyl.DifferentialActionModelLQR(2, 2))
    # multibody actions
    state = crocoddyl.StateMultibody(pinocchio.buildSampleModelHumanoidRandom())
    actuation = crocoddyl.ActuationModelFloatingBase(state)
    cost = crocoddyl.CostModelSum(state, actuation.nu)
    impulse = crocoddyl.ImpulseModelMultiple(state)
    contact = crocoddyl.ContactModelMultiple(state, actuation.nu)
    MODEL.append(crocoddyl.ActionModelImpulseFwdDynamics(state, impulse, cost))
    MODEL.append(
        crocoddyl.DifferentialActionModelContactFwdDynamics(
            state, actuation, contact, cost
        )
    )
    MODEL.append(
        crocoddyl.DifferentialActionModelFreeFwdDynamics(state, actuation, cost)
    )
    cost = crocoddyl.CostModelSum(state, state.nv)
    contact = crocoddyl.ContactModelMultiple(state, state.nv)
    MODEL.append(
        crocoddyl.DifferentialActionModelContactInvDynamics(
            state, actuation, contact, cost
        )
    )
    MODEL.append(
        crocoddyl.DifferentialActionModelFreeInvDynamics(state, actuation, cost)
    )
    # integrated actions
    MODEL.append(
        crocoddyl.IntegratedActionModelEuler(
            crocoddyl.DifferentialActionModelLQR(2, 2), 0.1
        )
    )
    MODEL.append(
        crocoddyl.IntegratedActionModelRK(
            crocoddyl.DifferentialActionModelLQR(2, 2), crocoddyl.RKType.two, 0.1
        )
    )
    # numdiff actions
    MODEL.append(crocoddyl.ActionModelNumDiff(crocoddyl.ActionModelLQR(2, 2)))
    MODEL.append(
        crocoddyl.DifferentialActionModelNumDiff(
            crocoddyl.DifferentialActionModelLQR(2, 2)
        )
    )


class StatesTest(CopyModelTestCase):
    MODEL = list()
    # core states
    MODEL.append(crocoddyl.StateVector(2))
    MODEL.append(crocoddyl.StateNumDiff(crocoddyl.StateVector(2)))
    # multibody states
    MODEL.append(crocoddyl.StateMultibody(pinocchio.buildSampleModelHumanoidRandom()))


class ResidualsTest(CopyModelTestCase):
    MODEL = list()
    DATA = True
    COLLECTOR = list()
    # core residuals
    state = crocoddyl.StateMultibody(pinocchio.buildSampleModelHumanoidRandom())
    actuation = crocoddyl.ActuationModelFloatingBase(state)
    joint = crocoddyl.JointDataAbstract(state, actuation, actuation.nu)
    MODEL.append(crocoddyl.ResidualModelControl(crocoddyl.StateVector(2)))
    COLLECTOR.append(crocoddyl.DataCollectorJoint(joint))
    MODEL.append(crocoddyl.ResidualModelJointEffort(state, actuation))
    COLLECTOR.append(crocoddyl.DataCollectorJoint(joint))
    MODEL.append(crocoddyl.ResidualModelJointAcceleration(state))
    COLLECTOR.append(crocoddyl.DataCollectorJoint(joint))
    # multibody residuals
    # TODO(cmastalli): add pair-collision residual
    frame_id = state.pinocchio.getFrameId("rleg6_joint")
    contact = crocoddyl.ContactModelMultiple(state, actuation.nu)
    contact.addContact(
        "rleg6_contact",
        crocoddyl.ContactModel6D(
            state,
            frame_id,
            pinocchio.SE3.Random(),
            pinocchio.LOCAL,
            actuation.nu,
            np.zeros(2),
        ),
    )
    pdata = state.pinocchio.createData()
    adata = actuation.createData()
    cdata = contact.createData(pdata)
    MODEL.append(crocoddyl.ResidualModelCentroidalMomentum(state, np.zeros(6)))
    COLLECTOR.append(crocoddyl.DataCollectorMultibody(pdata))
    MODEL.append(crocoddyl.ResidualModelCoMPosition(state, np.zeros(3)))
    COLLECTOR.append(crocoddyl.DataCollectorMultibody(pdata))
    MODEL.append(crocoddyl.ResidualModelControlGrav(state))
    COLLECTOR.append(crocoddyl.DataCollectorActMultibody(pdata, adata))
    MODEL.append(
        crocoddyl.ResidualModelFramePlacement(state, frame_id, pinocchio.SE3.Random())
    )
    COLLECTOR.append(crocoddyl.DataCollectorMultibody(pdata))
    MODEL.append(
        crocoddyl.ResidualModelFrameRotation(
            state, frame_id, pinocchio.SE3.Random().rotation
        )
    )
    COLLECTOR.append(crocoddyl.DataCollectorMultibody(pdata))
    MODEL.append(
        crocoddyl.ResidualModelFrameTranslation(
            state, frame_id, pinocchio.SE3.Random().translation
        )
    )
    COLLECTOR.append(crocoddyl.DataCollectorMultibody(pdata))
    MODEL.append(
        crocoddyl.ResidualModelFrameVelocity(
            state, frame_id, pinocchio.Motion.Random(), pinocchio.ReferenceFrame.LOCAL
        )
    )
    COLLECTOR.append(crocoddyl.DataCollectorMultibody(pdata))
    MODEL.append(crocoddyl.ResidualModelState(state))
    COLLECTOR.append(crocoddyl.DataCollectorAbstract())
    MODEL.append(crocoddyl.ResidualModelContactControlGrav(state))
    COLLECTOR.append(crocoddyl.DataCollectorActMultibodyInContact(pdata, adata, cdata))
    MODEL.append(
        crocoddyl.ResidualModelContactCoPPosition(
            state, frame_id, crocoddyl.CoPSupport()
        )
    )
    COLLECTOR.append(crocoddyl.DataCollectorActMultibodyInContact(pdata, adata, cdata))
    MODEL.append(
        crocoddyl.ResidualModelContactForce(
            state, frame_id, pinocchio.Force.Random(), 6
        )
    )
    COLLECTOR.append(crocoddyl.DataCollectorActMultibodyInContact(pdata, adata, cdata))
    MODEL.append(
        crocoddyl.ResidualModelContactFrictionCone(
            state, frame_id, crocoddyl.FrictionCone()
        )
    )
    COLLECTOR.append(crocoddyl.DataCollectorActMultibodyInContact(pdata, adata, cdata))
    MODEL.append(
        crocoddyl.ResidualModelContactWrenchCone(
            state, frame_id, crocoddyl.WrenchCone()
        )
    )
    COLLECTOR.append(crocoddyl.DataCollectorActMultibodyInContact(pdata, adata, cdata))


class ActivationsTest(CopyModelTestCase):
    MODEL = list()
    DATA = True
    bounds = crocoddyl.ActivationBounds(np.zeros(2), np.ones(2), 0.1)
    MODEL.append(crocoddyl.ActivationModel2NormBarrier(2))
    MODEL.append(crocoddyl.ActivationModelQuadraticBarrier(bounds))
    MODEL.append(crocoddyl.ActivationModelQuadFlatExp(2, 0.1))
    MODEL.append(crocoddyl.ActivationModelQuadFlatLog(2, 0.1))
    MODEL.append(crocoddyl.ActivationModelQuad(2))
    MODEL.append(crocoddyl.ActivationModelSmooth1Norm(2, 0.1))
    MODEL.append(crocoddyl.ActivationModelSmooth2Norm(2, 0.1))
    MODEL.append(crocoddyl.ActivationModelWeightedQuadraticBarrier(bounds, np.ones(2)))
    MODEL.append(crocoddyl.ActivationModelWeightedQuad(np.ones(2)))
    MODEL.append(
        crocoddyl.ActivationModelNumDiff(crocoddyl.ActivationModel2NormBarrier(2))
    )


class CostsTest(CopyModelTestCase):
    MODEL = list()
    DATA = True
    COLLECTOR = list()
    state = crocoddyl.StateVector(2)
    residual = crocoddyl.ResidualModelControl(crocoddyl.StateVector(2))
    activation = crocoddyl.ActivationModelWeightedQuad(np.ones(1))
    MODEL.append(crocoddyl.CostModelResidual(state, activation, residual))
    COLLECTOR.append(crocoddyl.DataCollectorAbstract())
    MODEL.append(crocoddyl.CostModelSum(state, 2))
    COLLECTOR.append(crocoddyl.DataCollectorAbstract())


class ConstraintsTest(CopyModelTestCase):
    MODEL = list()
    DATA = True
    COLLECTOR = list()
    state = crocoddyl.StateVector(2)
    residual = crocoddyl.ResidualModelControl(crocoddyl.StateVector(2))
    MODEL.append(crocoddyl.ConstraintModelResidual(state, residual))
    COLLECTOR.append(crocoddyl.DataCollectorAbstract())
    MODEL.append(crocoddyl.ConstraintModelManager(state, 2))
    COLLECTOR.append(crocoddyl.DataCollectorAbstract())


class ControlsTest(CopyModelTestCase):
    MODEL = list()
    DATA = True
    MODEL.append(crocoddyl.ControlParametrizationModelPolyZero(2))
    MODEL.append(crocoddyl.ControlParametrizationModelPolyOne(2))
    MODEL.append(
        crocoddyl.ControlParametrizationModelPolyTwoRK(2, crocoddyl.RKType.three)
    )


class DataCollectorsTest(CopyModelTestCase):
    MODEL = list()
    DATA = False
    # core collectors
    state = crocoddyl.StateMultibody(pinocchio.buildSampleModelHumanoidRandom())
    actuation = crocoddyl.ActuationModelFloatingBase(state)
    jdata = crocoddyl.JointDataAbstract(state, actuation, actuation.nu)
    MODEL.append(crocoddyl.DataCollectorAbstract())
    MODEL.append(crocoddyl.DataCollectorActuation(actuation.createData()))
    MODEL.append(jdata)
    MODEL.append(crocoddyl.DataCollectorJoint(jdata))
    # multibody collectors
    impulse = crocoddyl.ImpulseModelMultiple(state)
    contact = crocoddyl.ContactModelMultiple(state, actuation.nu)
    pdata = state.pinocchio.createData()
    adata = actuation.createData()
    cdata = contact.createData(pdata)
    idata = impulse.createData(pdata)
    MODEL.append(crocoddyl.DataCollectorMultibody(pdata))
    MODEL.append(crocoddyl.DataCollectorActMultibody(pdata, adata))
    MODEL.append(crocoddyl.DataCollectorJointActMultibody(pdata, adata, jdata))
    MODEL.append(crocoddyl.DataCollectorImpulse(idata))
    MODEL.append(crocoddyl.DataCollectorContact(cdata))
    MODEL.append(crocoddyl.DataCollectorMultibodyInImpulse(pdata, idata))
    MODEL.append(crocoddyl.DataCollectorMultibodyInContact(pdata, cdata))
    MODEL.append(crocoddyl.DataCollectorActMultibodyInContact(pdata, adata, cdata))
    MODEL.append(
        crocoddyl.DataCollectorJointActMultibodyInContact(pdata, adata, jdata, cdata)
    )
    cmodel = crocoddyl.ContactModelAbstract(state, pinocchio.LOCAL, 3, actuation.nu)
    MODEL.append(crocoddyl.ForceDataAbstract(cmodel, pdata))


class ActuationsTest(CopyModelTestCase):
    MODEL = list()
    DATA = True
    # core actuations
    state = crocoddyl.StateMultibody(pinocchio.buildSampleModelHumanoidRandom())
    actuation = crocoddyl.ActuationModelFloatingBase(state)
    MODEL.append(crocoddyl.SquashingModelSmoothSat(np.zeros(2), np.ones(2), 2))
    MODEL.append(
        crocoddyl.ActuationSquashingModel(
            actuation,
            crocoddyl.SquashingModelSmoothSat(np.zeros(2), np.ones(2), 2),
            actuation.nu,
        )
    )
    # multibody actuations
    MODEL.append(crocoddyl.ActuationModelFloatingBase(state))
    MODEL.append(crocoddyl.ActuationModelFull(state))
    d_cog, cf, cm, u_lim, l_lim = 0.1525, 6.6e-5, 1e-6, 5.0, 0.1
    ps = [
        crocoddyl.Thruster(
            pinocchio.SE3(np.eye(3), np.array([d_cog, 0, 0])),
            cm / cf,
            crocoddyl.ThrusterType.CCW,
        ),
        crocoddyl.Thruster(
            pinocchio.SE3(np.eye(3), np.array([0, d_cog, 0])),
            cm / cf,
            crocoddyl.ThrusterType.CW,
        ),
        crocoddyl.Thruster(
            pinocchio.SE3(np.eye(3), np.array([-d_cog, 0, 0])),
            cm / cf,
            crocoddyl.ThrusterType.CCW,
        ),
        crocoddyl.Thruster(
            pinocchio.SE3(np.eye(3), np.array([0, -d_cog, 0])),
            cm / cf,
            crocoddyl.ThrusterType.CW,
        ),
    ]
    actuation = crocoddyl.ActuationModelFloatingBaseThrusters(state, ps)
    MODEL.append(crocoddyl.ActuationModelFloatingBaseThrusters(state, ps))


class ContactsTest(CopyModelTestCase):
    MODEL = list()
    DATA = True
    COLLECTOR = list()
    state = crocoddyl.StateMultibody(pinocchio.buildSampleModelHumanoidRandom())
    actuation = crocoddyl.ActuationModelFloatingBase(state)
    frame_id = state.pinocchio.getFrameId("rleg6_joint")
    pdata = state.pinocchio.createData()
    # contact models
    MODEL.append(crocoddyl.ContactModelMultiple(state, actuation.nu))
    COLLECTOR.append(pdata)
    MODEL.append(
        crocoddyl.ContactModel1D(
            state, frame_id, 0.0, pinocchio.LOCAL, np.eye(3), actuation.nu, np.zeros(2)
        )
    )
    COLLECTOR.append(pdata)
    MODEL.append(
        crocoddyl.ContactModel2D(state, frame_id, np.ones(2), actuation.nu, np.zeros(2))
    )
    COLLECTOR.append(pdata)
    MODEL.append(
        crocoddyl.ContactModel3D(
            state, frame_id, np.ones(3), pinocchio.LOCAL, actuation.nu, np.zeros(2)
        )
    )
    COLLECTOR.append(pdata)
    MODEL.append(
        crocoddyl.ContactModel6D(
            state,
            frame_id,
            pinocchio.SE3.Random(),
            pinocchio.LOCAL,
            actuation.nu,
            np.zeros(2),
        )
    )
    COLLECTOR.append(pdata)
    # impulse models
    MODEL.append(crocoddyl.ImpulseModelMultiple(state))
    COLLECTOR.append(pdata)
    MODEL.append(crocoddyl.ImpulseModel3D(state, frame_id, pinocchio.LOCAL))
    COLLECTOR.append(pdata)
    MODEL.append(crocoddyl.ImpulseModel6D(state, frame_id))
    COLLECTOR.append(pdata)


class ConesTest(CopyModelTestCase):
    MODEL = list()
    MODEL.append(crocoddyl.FrictionCone())
    MODEL.append(crocoddyl.WrenchCone())
    MODEL.append(crocoddyl.CoPSupport())


class ProblemAndSolversTest(CopyModelTestCase):
    MODEL = list()
    m = crocoddyl.ActionModelLQR(2, 2)
    problem = crocoddyl.ShootingProblem(m.state.zero(), [m] * 10, m)
    MODEL.append(problem)
    MODEL.append(crocoddyl.CallbackVerbose())
    MODEL.append(crocoddyl.SolverKKT(problem))
    MODEL.append(crocoddyl.SolverDDP(problem))
    MODEL.append(crocoddyl.SolverFDDP(problem))
    MODEL.append(crocoddyl.SolverBoxDDP(problem))
    MODEL.append(crocoddyl.SolverBoxFDDP(problem))
    MODEL.append(crocoddyl.SolverIntro(problem))
    if hasattr(crocoddyl, "SolverIpopt"):
        MODEL.append(crocoddyl.SolverIpopt(problem))


if __name__ == "__main__":
    # test to be run
    test_classes_to_run = [
        ActionsTest,
        StatesTest,
        ResidualsTest,
        ActivationsTest,
        CostsTest,
        ConstraintsTest,
        ControlsTest,
        DataCollectorsTest,
        ActuationsTest,
        ContactsTest,
        ConesTest,
        ProblemAndSolversTest,
    ]
    loader = unittest.TestLoader()
    suites_list = []
    for test_class in test_classes_to_run:
        suite = loader.loadTestsFromTestCase(test_class)
        suites_list.append(suite)
    big_suite = unittest.TestSuite(suites_list)
    runner = unittest.TextTestRunner()
    results = runner.run(big_suite)
    sys.exit(not results.wasSuccessful())



import sys
import time

import example_robot_data
import numpy as np
import pinocchio

import crocoddyl

# First, let's load the Pinocchio model for the Talos arm.
ROBOT = example_robot_data.load("kinova")
N = 100  # number of nodes
T = int(sys.argv[1]) if (len(sys.argv) > 1) else int(5e3)  # number of trials
MAXITER = 1
CALLBACKS = False


def createProblem(model):
    robot_model = ROBOT.model
    q0 = robot_model.referenceConfigurations["arm_up"]
    x0 = np.concatenate([q0, np.zeros(robot_model.nv)])

    # Note that we need to include a cost model (i.e. set of cost functions) in
    # order to fully define the action model for our optimal control problem.
    # For this particular example, we formulate three running-cost functions:
    # goal-tracking cost, state and control regularization; and one terminal-cost:
    # goal cost. First, let's create the common cost functions.
    state = crocoddyl.StateMultibody(robot_model)
    goalTrackingCost = crocoddyl.CostModelResidual(
        state,
        crocoddyl.ResidualModelFramePlacement(
            state,
            robot_model.getFrameId("j2s6s200_end_effector"),
            pinocchio.SE3(np.eye(3), np.array([0.6, 0.2, 0.5])),
        ),
    )
    xRegCost = crocoddyl.CostModelResidual(state, crocoddyl.ResidualModelState(state))
    uRegCost = crocoddyl.CostModelResidual(state, crocoddyl.ResidualModelControl(state))

    # Create a cost model per the running and terminal action model.
    runningCostModel = crocoddyl.CostModelSum(state)
    terminalCostModel = crocoddyl.CostModelSum(state)

    # Then let's added the running and terminal cost functions
    runningCostModel.addCost("gripperPose", goalTrackingCost, 1)
    runningCostModel.addCost("xReg", xRegCost, 1e-1)
    runningCostModel.addCost("uReg", uRegCost, 1e-1)
    terminalCostModel.addCost("gripperPose", goalTrackingCost, 1e3)

    # Next, we need to create an action model for running and terminal knots. The
    # forward dynamics (computed using ABA) are implemented
    # inside DifferentialActionModelFullyActuated.
    actuation = crocoddyl.ActuationModelFull(state)
    runningModel = crocoddyl.IntegratedActionModelEuler(
        model(state, actuation, runningCostModel), 1e-2
    )
    terminalModel = crocoddyl.IntegratedActionModelEuler(
        model(state, actuation, terminalCostModel), 0.0
    )

    # For this optimal control problem, we define 100 knots (or running action
    # models) plus a terminal knot
    problem = crocoddyl.ShootingProblem(x0, [runningModel] * N, terminalModel)
    xs = [x0] * (len(problem.runningModels) + 1)
    us = [
        m.quasiStatic(d, x0)
        for m, d in list(zip(problem.runningModels, problem.runningDatas))
    ]
    return xs, us, problem


def runDDPSolveBenchmark(xs, us, problem):
    solver = crocoddyl.SolverFDDP(problem)
    if CALLBACKS:
        solver.setCallbacks([crocoddyl.CallbackVerbose()])
    duration = []
    for _ in range(T):
        c_start = time.time()
        solver.solve(xs, us, MAXITER, False, 0.1)
        c_end = time.time()
        duration.append(1e3 * (c_end - c_start))

    avrg_duration = sum(duration) / len(duration)
    min_duration = min(duration)
    max_duration = max(duration)
    return avrg_duration, min_duration, max_duration


def runShootingProblemCalcBenchmark(xs, us, problem):
    duration = []
    for _ in range(T):
        c_start = time.time()
        problem.calc(xs, us)
        c_end = time.time()
        duration.append(1e3 * (c_end - c_start))

    avrg_duration = sum(duration) / len(duration)
    min_duration = min(duration)
    max_duration = max(duration)
    return avrg_duration, min_duration, max_duration


def runShootingProblemCalcDiffBenchmark(xs, us, problem):
    duration = []
    for _ in range(T):
        c_start = time.time()
        problem.calcDiff(xs, us)
        c_end = time.time()
        duration.append(1e3 * (c_end - c_start))

    avrg_duration = sum(duration) / len(duration)
    min_duration = min(duration)
    max_duration = max(duration)
    return avrg_duration, min_duration, max_duration


np.set_printoptions(precision=2)
xs, us, problem = createProblem(crocoddyl.DifferentialActionModelFreeFwdDynamics)
print("NQ:", problem.terminalModel.state.nq)
print("Number of nodes:", problem.T)
avrg_dur, min_dur, max_dur = runDDPSolveBenchmark(xs, us, problem)
print(f"  FDDP.solve [ms]: {avrg_dur:.4f} ({min_dur:.4f}-{max_dur:.4f})")
avrg_dur, min_dur, max_dur = runShootingProblemCalcBenchmark(xs, us, problem)
print(f"  ShootingProblem.calc [ms]: {avrg_dur:.4f} ({min_dur:.4f}-{max_dur:.4f})")
avrg_dur, min_dur, max_dur = runShootingProblemCalcDiffBenchmark(xs, us, problem)
print(f"  ShootingProblem.calcDiff [ms]: {avrg_dur:.4f} ({min_dur:.4f}-{max_dur:.4f})")



# flake8: noqa: F821
# ##########################################################################
# ################# TODO: Tune the weights for each cost ###################
# ##########################################################################
terminalCartpole = DifferentialActionModelCartpole()
terminalCartpoleDAM = crocoddyl.DifferentialActionModelNumDiff(terminalCartpole, True)
terminalCartpoleIAM = crocoddyl.IntegratedActionModelEuler(terminalCartpoleDAM)

terminalCartpole.costWeights[0] = 100
terminalCartpole.costWeights[1] = 100
terminalCartpole.costWeights[2] = 1.0
terminalCartpole.costWeights[3] = 0.1
terminalCartpole.costWeights[4] = 0.01
terminalCartpole.costWeights[5] = 0.0001
problem = crocoddyl.ShootingProblem(x0, [cartpoleIAM] * T, terminalCartpoleIAM)



import sys
import time

import numpy as np

import crocoddyl

NX = 37
NU = 12
N = 100  # number of nodes
T = int(sys.argv[1]) if (len(sys.argv) > 1) else int(5e3)  # number of trials
MAXITER = 1
CALLBACKS = False


def createProblem(model):
    x0 = np.matrix(np.zeros(NX)).T
    runningModels = [model(NX, NU)] * N
    terminalModel = model(NX, NU)
    xs = [x0] * (N + 1)
    us = [np.matrix(np.zeros(NU)).T] * N

    problem = crocoddyl.ShootingProblem(x0, runningModels, terminalModel)
    return xs, us, problem


def runDDPSolveBenchmark(xs, us, problem):
    solver = crocoddyl.SolverFDDP(problem)
    if CALLBACKS:
        solver.setCallbacks([crocoddyl.CallbackVerbose()])
    duration = []
    for i in range(T):
        c_start = time.time()
        solver.solve(xs, us, MAXITER)
        c_end = time.time()
        duration.append(1e3 * (c_end - c_start))

    avrg_duration = sum(duration) / len(duration)
    min_duration = min(duration)
    max_duration = max(duration)
    return avrg_duration, min_duration, max_duration


def runShootingProblemCalcBenchmark(xs, us, problem):
    duration = []
    for i in range(T):
        c_start = time.time()
        problem.calc(xs, us)
        c_end = time.time()
        duration.append(1e3 * (c_end - c_start))

    avrg_duration = sum(duration) / len(duration)
    min_duration = min(duration)
    max_duration = max(duration)
    return avrg_duration, min_duration, max_duration


def runShootingProblemCalcDiffBenchmark(xs, us, problem):
    duration = []
    for i in range(T):
        c_start = time.time()
        problem.calcDiff(xs, us)
        c_end = time.time()
        duration.append(1e3 * (c_end - c_start))

    avrg_duration = sum(duration) / len(duration)
    min_duration = min(duration)
    max_duration = max(duration)
    return avrg_duration, min_duration, max_duration


xs, us, problem = createProblem(crocoddyl.ActionModelLQR)
print("NQ:", problem.terminalModel.state.nq)
print("Number of nodes:", problem.T)
avrg_dur, min_dur, max_dur = runDDPSolveBenchmark(xs, us, problem)
print(f"  FDDP.solve [ms]: {avrg_dur:.4f} ({min_dur:.4f}-{max_dur:.4f})")
avrg_dur, min_dur, max_dur = runShootingProblemCalcBenchmark(xs, us, problem)
print(f"  ShootingProblem.calc [ms]: {avrg_dur:.4f} ({min_dur:.4f}-{max_dur:.4f})")
avrg_dur, min_dur, max_dur = runShootingProblemCalcDiffBenchmark(xs, us, problem)
print(f"  ShootingProblem.calcDiff [ms]: {avrg_dur:.4f} ({min_dur:.4f}-{max_dur:.4f})")



import collections
import sys
import unittest

import example_robot_data
import numpy as np
import pinocchio
from factory import Impulse3DModelDerived, Impulse6DModelDerived

import crocoddyl


class ImpulseModelAbstractTestCase(unittest.TestCase):
    ROBOT_MODEL = None
    ROBOT_STATE = None
    IMPULSE = None
    IMPULSE_DER = None

    def setUp(self):
        self.x = self.ROBOT_STATE.rand()
        self.robot_data = self.ROBOT_MODEL.createData()
        self.data = self.IMPULSE.createData(self.robot_data)
        self.data_der = self.IMPULSE_DER.createData(self.robot_data)

        nq, nv = self.ROBOT_MODEL.nq, self.ROBOT_MODEL.nv
        pinocchio.forwardKinematics(
            self.ROBOT_MODEL,
            self.robot_data,
            self.x[:nq],
            self.x[nq:],
            pinocchio.utils.zero(nv),
        )
        pinocchio.computeJointJacobians(self.ROBOT_MODEL, self.robot_data)
        pinocchio.updateFramePlacements(self.ROBOT_MODEL, self.robot_data)
        pinocchio.computeForwardKinematicsDerivatives(
            self.ROBOT_MODEL,
            self.robot_data,
            self.x[:nq],
            self.x[nq:],
            pinocchio.utils.zero(nv),
        )

    def test_nc_dimension(self):
        self.assertEqual(self.IMPULSE.nc, self.IMPULSE_DER.nc, "Wrong ni.")

    def test_calc(self):
        # Run calc for both action models
        self.IMPULSE.calc(self.data, self.x)
        self.IMPULSE_DER.calc(self.data_der, self.x)
        # Checking the cost value and its residual
        self.assertTrue(
            np.allclose(self.data.Jc, self.data_der.Jc, atol=1e-9),
            "Wrong contact Jacobian (Jc).",
        )

    def test_calcDiff(self):
        # Run calc for both action models
        self.IMPULSE.calc(self.data, self.x)
        self.IMPULSE.calcDiff(self.data, self.x)

        self.IMPULSE_DER.calc(self.data_der, self.x)
        self.IMPULSE_DER.calcDiff(self.data_der, self.x)
        # Checking the Jacobians of the contact constraint
        self.assertTrue(
            np.allclose(self.data.dv0_dq, self.data_der.dv0_dq, atol=1e-9),
            "Wrong Jacobian of the acceleration before impulse (dv0_dq).",
        )

    def test_updateForce(self):
        # Run updateForce for both action models
        self.IMPULSE.calc(self.data, self.x)
        self.IMPULSE_DER.calc(self.data_der, self.x)
        rng = np.random.default_rng()
        force = rng.random(self.IMPULSE.nc)
        self.IMPULSE.updateForce(self.data, force)
        self.IMPULSE_DER.updateForce(self.data_der, force)
        self.assertTrue(
            np.allclose(self.data.fext.vector, self.data_der.fext.vector, atol=1e-9),
            "Wrong external spatial force.",
        )


class ImpulseModelMultipleAbstractTestCase(unittest.TestCase):
    ROBOT_MODEL = None
    ROBOT_STATE = None
    IMPULSES = None

    def setUp(self):
        self.x = self.ROBOT_STATE.rand()
        self.robot_data = self.ROBOT_MODEL.createData()

        self.impulseSum = crocoddyl.ImpulseModelMultiple(self.ROBOT_STATE)
        self.datas = collections.OrderedDict(
            [
                [name, impulse.createData(self.robot_data)]
                for name, impulse in self.IMPULSES.items()
            ]
        )
        for name, impulse in self.IMPULSES.items():
            self.impulseSum.addImpulse(name, impulse)
        self.dataSum = self.impulseSum.createData(self.robot_data)

        nq, nv = self.ROBOT_MODEL.nq, self.ROBOT_MODEL.nv
        pinocchio.forwardKinematics(
            self.ROBOT_MODEL,
            self.robot_data,
            self.x[:nq],
            self.x[nq:],
            pinocchio.utils.zero(nv),
        )
        pinocchio.computeJointJacobians(self.ROBOT_MODEL, self.robot_data)
        pinocchio.updateFramePlacements(self.ROBOT_MODEL, self.robot_data)
        pinocchio.computeForwardKinematicsDerivatives(
            self.ROBOT_MODEL,
            self.robot_data,
            self.x[:nq],
            self.x[nq:],
            pinocchio.utils.zero(nv),
        )

    def test_ni_dimension(self):
        nc = sum([impulse.nc for impulse in self.IMPULSES.values()])
        self.assertEqual(self.impulseSum.nc, nc, "Wrong nc.")

    def test_calc(self):
        # Run calc for both action models
        for impulse, data in zip(self.IMPULSES.values(), self.datas.values()):
            impulse.calc(data, self.x)
        self.impulseSum.calc(self.dataSum, self.x)
        # Checking the cost value and its residual
        Jc = np.vstack([data.Jc for data in self.datas.values()])
        self.assertTrue(
            np.allclose(self.dataSum.Jc, Jc, atol=1e-9), "Wrong contact Jacobian (Jc)."
        )

    def test_calcDiff(self):
        # Run calc for both action models
        for impulse, data in zip(self.IMPULSES.values(), self.datas.values()):
            impulse.calc(data, self.x)
            impulse.calcDiff(data, self.x)
        self.impulseSum.calc(self.dataSum, self.x)
        self.impulseSum.calcDiff(self.dataSum, self.x)
        # Checking the Jacobians of the contact constraint
        dv0_dq = np.vstack([data.dv0_dq for data in self.datas.values()])
        self.assertTrue(
            np.allclose(self.dataSum.dv0_dq, dv0_dq, atol=1e-9),
            "Wrong Jacobian of the velocity before impulse (dv0_dq).",
        )


class Impulse3DLocalTest(ImpulseModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("hyq").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    # gains = pinocchio.utils.rand(2)
    frame = ROBOT_MODEL.getFrameId("lf_foot")
    IMPULSE = crocoddyl.ImpulseModel3D(ROBOT_STATE, frame, pinocchio.LOCAL)
    IMPULSE_DER = Impulse3DModelDerived(ROBOT_STATE, frame, pinocchio.LOCAL)


class Impulse3DWorldTest(ImpulseModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("hyq").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    # gains = pinocchio.utils.rand(2)
    frame = ROBOT_MODEL.getFrameId("lf_foot")
    IMPULSE = crocoddyl.ImpulseModel3D(ROBOT_STATE, frame, pinocchio.WORLD)
    IMPULSE_DER = Impulse3DModelDerived(ROBOT_STATE, frame, pinocchio.WORLD)


class Impulse3DLocalWorldAlignedTest(ImpulseModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("hyq").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    # gains = pinocchio.utils.rand(2)
    frame = ROBOT_MODEL.getFrameId("lf_foot")
    IMPULSE = crocoddyl.ImpulseModel3D(
        ROBOT_STATE, frame, pinocchio.LOCAL_WORLD_ALIGNED
    )
    IMPULSE_DER = Impulse3DModelDerived(
        ROBOT_STATE, frame, pinocchio.LOCAL_WORLD_ALIGNED
    )


class Impulse3DMultipleTest(ImpulseModelMultipleAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("hyq").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    gains = pinocchio.utils.rand(2)
    IMPULSES = collections.OrderedDict(
        sorted(
            {
                "lf_foot": crocoddyl.ImpulseModel3D(
                    ROBOT_STATE, ROBOT_MODEL.getFrameId("lf_foot"), pinocchio.LOCAL
                ),
                "rh_foot": crocoddyl.ImpulseModel3D(
                    ROBOT_STATE, ROBOT_MODEL.getFrameId("rh_foot"), pinocchio.LOCAL
                ),
            }.items()
        )
    )


class Impulse6DLocalTest(ImpulseModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("icub_reduced").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    frame = ROBOT_MODEL.getFrameId("r_sole")
    IMPULSE = crocoddyl.ImpulseModel6D(ROBOT_STATE, frame, pinocchio.LOCAL)
    IMPULSE_DER = Impulse6DModelDerived(ROBOT_STATE, frame, pinocchio.LOCAL)


class Impulse6DWorldTest(ImpulseModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("icub_reduced").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    frame = ROBOT_MODEL.getFrameId("r_sole")
    IMPULSE = crocoddyl.ImpulseModel6D(ROBOT_STATE, frame, pinocchio.WORLD)
    IMPULSE_DER = Impulse6DModelDerived(ROBOT_STATE, frame, pinocchio.WORLD)


class Impulse6DLocalWorldAlignedTest(ImpulseModelAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("icub_reduced").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    frame = ROBOT_MODEL.getFrameId("r_sole")
    IMPULSE = crocoddyl.ImpulseModel6D(
        ROBOT_STATE, frame, pinocchio.LOCAL_WORLD_ALIGNED
    )
    IMPULSE_DER = Impulse6DModelDerived(
        ROBOT_STATE, frame, pinocchio.LOCAL_WORLD_ALIGNED
    )


class Impulse6DMultipleTest(ImpulseModelMultipleAbstractTestCase):
    ROBOT_MODEL = example_robot_data.load("icub_reduced").model
    ROBOT_STATE = crocoddyl.StateMultibody(ROBOT_MODEL)

    gains = pinocchio.utils.rand(2)
    IMPULSES = collections.OrderedDict(
        sorted(
            {
                "l_sole": crocoddyl.ImpulseModel6D(
                    ROBOT_STATE, ROBOT_MODEL.getFrameId("l_sole")
                ),
                "r_sole": crocoddyl.ImpulseModel6D(
                    ROBOT_STATE, ROBOT_MODEL.getFrameId("r_sole")
                ),
            }.items()
        )
    )


if __name__ == "__main__":
    # test to be run
    test_classes_to_run = [
        Impulse3DLocalTest,
        Impulse3DWorldTest,
        Impulse3DLocalWorldAlignedTest,
        Impulse3DMultipleTest,
        Impulse6DLocalTest,
        Impulse6DWorldTest,
        Impulse6DLocalWorldAlignedTest,
        Impulse6DMultipleTest,
    ]
    loader = unittest.TestLoader()
    suites_list = []
    for test_class in test_classes_to_run:
        suite = loader.loadTestsFromTestCase(test_class)
        suites_list.append(suite)
    big_suite = unittest.TestSuite(suites_list)
    runner = unittest.TextTestRunner()
    results = runner.run(big_suite)
    sys.exit(not results.wasSuccessful())



import example_robot_data
import numpy as np
import pinocchio

import crocoddyl

robot = example_robot_data.load("talos_arm")
robot_model = robot.model

DT = 1e-3
T = 150

state = crocoddyl.StateMultibody(robot.model)

ps = [
    np.array([0.4, 0.0, 0.4]),
    np.array([0.4, 0.4, 0.4]),
    np.array([0.4, 0.4, 0.0]),
    np.array([0.4, 0.0, 0.0]),
]

colors = [
    [1.0, 0.0, 0.0, 1.0],
    [0.0, 1.0, 0.0, 1.0],
    [0.0, 0.0, 1.0, 1.0],
    [1.0, 0.0, 1.0, 1.0],
]

cameraTF = [2.0, 2.68, 0.54, 0.2, 0.62, 0.72, 0.22]
display = crocoddyl.GepettoDisplay(robot, cameraTF=cameraTF, floor=False)
gv = display.robot.viewer.gui
for i, p in enumerate(ps):
    gv.addSphere("world/point%d" % i, 0.05, colors[i])
    gv.applyConfiguration("world/point%d" % i, [*p.tolist(), 0.0, 0.0, 0.0, 1.0])
gv.refresh()

# State and control regularization costs
xRegCost = crocoddyl.CostModelResidual(state, crocoddyl.ResidualModelState(state))
uRegCost = crocoddyl.CostModelResidual(state, crocoddyl.ResidualModelControl(state))

# Then let's added the running and terminal cost functions per each action
# model
runningModels = []
terminalModels = []
for p in ps:
    # Create the tracking cost
    goalTrackingCost = crocoddyl.CostModelResidual(
        state,
        crocoddyl.ResidualFrameTranslation(
            state, robot_model.getFrameId("gripper_left_joint"), p
        ),
    )

    actuation = crocoddyl.ActuationModelFull(state)

    # Create the running action model
    runningCostModel = crocoddyl.CostModelSum(state)
    runningCostModel.addCost("gripperPose", goalTrackingCost, 1e-3)
    runningCostModel.addCost("xReg", xRegCost, 1e-4)
    runningCostModel.addCost("uReg", uRegCost, 1e-7)
    runningModels += [
        crocoddyl.DifferentialActionModelFreeFwdDynamics(
            state, actuation, runningCostModel
        )
    ]

    # Create the terminal action model
    terminalCostModel = crocoddyl.CostModelSum(state)
    terminalCostModel.addCost("gripperPose", goalTrackingCost, 1.0)
    terminalCostModel.addCost("xreg", xRegCost, 1e-4)
    terminalCostModel.addCost("uReg", uRegCost, 1e-7)
    terminalModels += [
        crocoddyl.DifferentialActionModelFreeFwdDynamics(
            state, actuation, terminalCostModel
        )
    ]

q0 = np.array([2.0, 1.5, -2.0, 0.0, 0.0, 0.0, 0.0]).T
x0 = np.concatenate([q0, pinocchio.utils.zero(state.nv)])
seqs = [
    [crocoddyl.IntegratedActionModelEuler(runningModel, DT)] * T
    + [crocoddyl.IntegratedActionModelEuler(terminalModel)]
    for runningModel, terminalModel in zip(runningModels, terminalModels)
]
problem = crocoddyl.ShootingProblem(
    x0,
    sum(seqs, [])[:-1],  # noqa: RUF017
    seqs[-1][-1],
)

# Creating the DDP solver for this OC problem, defining a logger
ddp = crocoddyl.SolverDDP(problem)
ddp.setCallbacks([crocoddyl.CallbackVerbose()])

# Solving it with the DDP algorithm
ddp.solve()

# Penalty if you want.
for i in range(4, 6):
    for m in terminalModels:
        m.costs.costs["gripperPose"].weight = 10**i
    ddp.solve(ddp.xs, ddp.us, 10)

# Visualizing the solution in gepetto-viewer
display.displayFromSolver(ddp)

robot_data = robot_model.createData()
for i, s in enumerate([*seqs, 1]):
    xT = ddp.xs[i * T]
    pinocchio.forwardKinematics(robot_model, robot_data, xT[: state.nq])
    pinocchio.updateFramePlacements(robot_model, robot_data)
    print(
        "Finally reached = ",
        robot_data.oMf[robot_model.getFrameId("gripper_left_joint")].translation.T,
    )



import os
import signal
import sys
import time

import example_robot_data
import numpy as np
import pinocchio

import crocoddyl

WITHDISPLAY = "display" in sys.argv or "CROCODDYL_DISPLAY" in os.environ
WITHPLOT = "plot" in sys.argv or "CROCODDYL_PLOT" in os.environ
signal.signal(signal.SIGINT, signal.SIG_DFL)

hector = example_robot_data.load("hector")
robot_model = hector.model

target_pos = np.array([1.0, 0.0, 1.0])
target_quat = pinocchio.Quaternion(1.0, 0.0, 0.0, 0.0)

state = crocoddyl.StateMultibody(robot_model)

d_cog, cf, cm, u_lim, l_lim = 0.1525, 6.6e-5, 1e-6, 5.0, 0.1
ps = [
    crocoddyl.Thruster(
        pinocchio.SE3(np.eye(3), np.array([d_cog, 0, 0])),
        cm / cf,
        crocoddyl.ThrusterType.CCW,
    ),
    crocoddyl.Thruster(
        pinocchio.SE3(np.eye(3), np.array([0, d_cog, 0])),
        cm / cf,
        crocoddyl.ThrusterType.CW,
    ),
    crocoddyl.Thruster(
        pinocchio.SE3(np.eye(3), np.array([-d_cog, 0, 0])),
        cm / cf,
        crocoddyl.ThrusterType.CCW,
    ),
    crocoddyl.Thruster(
        pinocchio.SE3(np.eye(3), np.array([0, -d_cog, 0])),
        cm / cf,
        crocoddyl.ThrusterType.CW,
    ),
]
actuation = crocoddyl.ActuationModelFloatingBaseThrusters(state, ps)

nu = actuation.nu
runningCostModel = crocoddyl.CostModelSum(state, nu)
terminalCostModel = crocoddyl.CostModelSum(state, nu)

# Costs
xResidual = crocoddyl.ResidualModelState(state, state.zero(), nu)
xActivation = crocoddyl.ActivationModelWeightedQuad(
    np.array([0.1] * 3 + [1000.0] * 3 + [1000.0] * robot_model.nv)
)
uResidual = crocoddyl.ResidualModelControl(state, nu)
xRegCost = crocoddyl.CostModelResidual(state, xActivation, xResidual)
uRegCost = crocoddyl.CostModelResidual(state, uResidual)
goalTrackingResidual = crocoddyl.ResidualModelFramePlacement(
    state,
    robot_model.getFrameId("base_link"),
    pinocchio.SE3(target_quat.matrix(), target_pos),
    nu,
)
goalTrackingCost = crocoddyl.CostModelResidual(state, goalTrackingResidual)
runningCostModel.addCost("xReg", xRegCost, 1e-6)
runningCostModel.addCost("uReg", uRegCost, 1e-6)
runningCostModel.addCost("trackPose", goalTrackingCost, 1e-2)
terminalCostModel.addCost("goalPose", goalTrackingCost, 3.0)

dt = 3e-2
runningModel = crocoddyl.IntegratedActionModelEuler(
    crocoddyl.DifferentialActionModelFreeFwdDynamics(
        state, actuation, runningCostModel
    ),
    dt,
)
terminalModel = crocoddyl.IntegratedActionModelEuler(
    crocoddyl.DifferentialActionModelFreeFwdDynamics(
        state, actuation, terminalCostModel
    ),
    dt,
)
runningModel.u_lb = np.array([l_lim, l_lim, l_lim, l_lim])
runningModel.u_ub = np.array([u_lim, u_lim, u_lim, u_lim])

# Creating the shooting problem and the BoxDDP solver
T = 33
problem = crocoddyl.ShootingProblem(
    np.concatenate([hector.q0, np.zeros(state.nv)]), [runningModel] * T, terminalModel
)
solver = crocoddyl.SolverBoxDDP(problem)
if WITHPLOT:
    solver.setCallbacks(
        [
            crocoddyl.CallbackVerbose(),
            crocoddyl.CallbackLogger(),
        ]
    )
else:
    solver.setCallbacks([crocoddyl.CallbackVerbose()])

# Solving the problem with the BoxDDP solver
solver.solve([], [], 200)

# Plotting the entire motion
if WITHPLOT:
    log = solver.getCallbacks()[1]
    crocoddyl.plotOCSolution(log.xs, log.us, figIndex=1, show=False)
    crocoddyl.plotConvergence(
        log.costs, log.pregs, log.dregs, log.grads, log.stops, log.steps, figIndex=2
    )

# Display the entire motion
if WITHDISPLAY:
    try:
        import gepetto

        gepetto.corbaserver.Client()
        cameraTF = [-0.03, 4.4, 2.3, -0.02, 0.56, 0.83, -0.03]
        display = crocoddyl.GepettoDisplay(hector, 4, 4, cameraTF, floor=False)
        hector.viewer.gui.addXYZaxis("world/wp", [1.0, 0.0, 0.0, 1.0], 0.03, 0.5)
        hector.viewer.gui.applyConfiguration(
            "world/wp",
            [
                *target_pos.tolist(),
                target_quat[0],
                target_quat[1],
                target_quat[2],
                target_quat[3],
            ],
        )
    except Exception:
        display = crocoddyl.MeshcatDisplay(hector)
    display.rate = -1
    display.freq = 1
    while True:
        display.displayFromSolver(solver)
        time.sleep(1.0)



# flake8: noqa: F821
# Use this function inside DifferentialActionModel.calc by setting:
#      xddot, thddot = cartpole_dynamics(self, data, x, u)


def cartpole_dynamics(model, data, x, u):
    # Getting the state and control variables
    th, thdot = x[1].item(), x[3].item()
    f = u[0].item()

    # Shortname for system parameters
    m1, m2, le, g = model.m1, model.m2, model.l, model.g
    s, c = np.sin(th), np.cos(th)

    # Defining the equation of motions
    m = m1 + m2
    mu = m1 + m2 * s**2
    xddot = (f + m2 * c * s * g - m2 * le * s * thdot) / mu
    thddot = (c * f / le + m * g * s / le - m2 * c * s * thdot**2) / mu

    return [xddot, thddot]



import os
import signal
import sys
import time

import example_robot_data
import numpy as np
import pinocchio

import crocoddyl
from crocoddyl.utils.quadruped import SimpleQuadrupedalGaitProblem, plotSolution

WITHDISPLAY = "display" in sys.argv or "CROCODDYL_DISPLAY" in os.environ
WITHPLOT = "plot" in sys.argv or "CROCODDYL_PLOT" in os.environ
signal.signal(signal.SIGINT, signal.SIG_DFL)

# Loading the anymal model
anymal = example_robot_data.load("anymal")
robot_model = anymal.model
lims = robot_model.effortLimit
lims *= 0.5  # reduced artificially the torque limits
robot_model.effortLimit = lims

# Setting up the 3d walking problem
lfFoot, rfFoot, lhFoot, rhFoot = "LF_FOOT", "RF_FOOT", "LH_FOOT", "RH_FOOT"
gait = SimpleQuadrupedalGaitProblem(robot_model, lfFoot, rfFoot, lhFoot, rhFoot)

# Defining the initial state of the robot
q0 = robot_model.referenceConfigurations["standing"].copy()
v0 = pinocchio.utils.zero(robot_model.nv)
x0 = np.concatenate([q0, v0])

# Defining the walking gait parameters
walking_gait = {
    "stepLength": 0.25,
    "stepHeight": 0.25,
    "timeStep": 1e-2,
    "stepKnots": 25,
    "supportKnots": 2,
}

# Setting up the control-limited DDP solver
solver = crocoddyl.SolverBoxDDP(
    gait.createWalkingProblem(
        x0,
        walking_gait["stepLength"],
        walking_gait["stepHeight"],
        walking_gait["timeStep"],
        walking_gait["stepKnots"],
        walking_gait["supportKnots"],
    )
)

# Add the callback functions
print("*** SOLVE ***")
if WITHPLOT:
    solver.setCallbacks(
        [
            crocoddyl.CallbackVerbose(),
            crocoddyl.CallbackLogger(),
        ]
    )
else:
    solver.setCallbacks([crocoddyl.CallbackVerbose()])

# Solve the DDP problem
xs = [x0] * (solver.problem.T + 1)
us = solver.problem.quasiStatic([x0] * solver.problem.T)
solver.solve(xs, us, 100, False, 0.1)

# Plotting the entire motion
if WITHPLOT:
    # Plot control vs limits
    plotSolution(solver, bounds=True, figIndex=1, show=False)

    # Plot convergence
    log = solver.getCallbacks()[1]
    crocoddyl.plotConvergence(
        log.costs,
        log.pregs,
        log.dregs,
        log.grads,
        log.stops,
        log.steps,
        figIndex=3,
        show=True,
    )

# Display the entire motion
if WITHDISPLAY:
    try:
        import gepetto

        gepetto.corbaserver.Client()
        cameraTF = [2.0, 2.68, 0.84, 0.2, 0.62, 0.72, 0.22]
        display = crocoddyl.GepettoDisplay(anymal, 4, 4, cameraTF)
    except Exception:
        display = crocoddyl.MeshcatDisplay(anymal)
    display.rate = -1
    display.freq = 1
    while True:
        display.displayFromSolver(solver)
        time.sleep(1.0)
